
import collections
import pandas as pd
import numpy as np
import scipy
import scipy.stats

import destruct.utils.misc
import destruct.utils.streaming


cluster_fields = ['cluster_id', 'cluster_end',
                  'library_id', 'read_id', 'read_end', 'align_id']


spanning_fields = ['library_id', 'read_id', 'read_end', 'align_id',
                   'chromosome', 'strand', 'position', 'aligned_length',
                   'mate_length', 'mate_score']


split_fields = ['library_id', 'read_id', 'read_end',
                'align_id_1', 'chromosome_1', 'strand_1', 'position_1',
                'align_id_2', 'chromosome_2', 'strand_2', 'position_2',
                'homology', 'inserted', 'score']


breakpoint_fields = ['cluster_id', 'breakpoint_id',
                     'chromosome_1', 'strand_1', 'position_1',
                     'chromosome_2', 'strand_2', 'position_2',
                     'homology', 'count', 'inserted',
                     'mate_score']


realignment_fields = ['cluster_id', 'breakpoint_id', 'cluster_end',
                      'library_id', 'read_id', 'read_end', 'align_id',
                      'aligned_length', 'template_length', 'score']


score_stats_fields = ['aligned_length', 'expon_lda']


likelihoods_fields = ['cluster_id', 'breakpoint_id',
                      'library_id', 'read_id',
                      'read_end_1', 'read_end_2',
                      'aligned_length_1', 'aligned_length_2',
                      'template_length_1', 'template_length_2',
                      'score_1', 'score_2',
                      'score_log_likelihood_1', 'score_log_likelihood_2',
                      'score_log_cdf_1', 'score_log_cdf_2',
                      'inslen', 'template_length', 'length_z_score',
                      'length_log_likelihood', 'length_log_cdf',
                      'log_likelihood', 'log_cdf']


def calculate_mate_score(clusters, spanning):

    # Create a table of spanning alignments for these clusters
    data = spanning.merge(clusters, on=['library_id', 'read_id', 'read_end', 'align_id'])

    data = data.groupby(['cluster_id', 'library_id', 'read_id'])[['mate_score']].max()\
               .groupby(level=0).mean()\
               .reset_index()

    return data


def predict_breaks_spanning(clusters, spanning):

    # Create a table of spanning alignments for these clusters
    data = spanning.merge(clusters, on=['library_id', 'read_id', 'read_end', 'align_id'])

    # Add start and end based on position and aligned length
    data['start'] = np.where(data['strand'] == '+', data['position'], data['position'] - data['aligned_length'] + 1)
    data['end'] = np.where(data['strand'] == '+', data['position'] + data['aligned_length'] - 1, data['position'])

    # Predict based on spanning reads
    agg_f = {'chromosome':max, 'strand':max, 'start':min, 'end':max}
    data = data.groupby(['cluster_id', 'cluster_end']).agg(agg_f).reset_index()
    data['position'] = np.where(data['strand'] == '+', data['end'], data['start'])
    data = data.drop(['start', 'end'], axis=1)

    # Reformat table
    data = data.set_index(['cluster_id', 'cluster_end']).unstack()
    data.columns = [a+'_'+str(b+1) for a, b in data.columns.values]
    data.reset_index(inplace=True)
    data['count'] = 0
    data['homology'] = 0
    data['inserted'] = ''
    data['breakpoint_id'] = 0

    return data


def predict_breaks_split(clusters, split, max_predictions_per_cluster=10):

    # Unstack clusters file by read end, in preparation for
    # merge with split alignments
    index_fields = ['cluster_id', 'library_id', 'read_id']
    data_fields = ['cluster_end', 'align_id']
    paired = pd.merge(clusters.loc[clusters['read_end'] == 0, index_fields + data_fields].drop_duplicates(),
                      clusters.loc[clusters['read_end'] == 1, index_fields + data_fields].drop_duplicates(),
                      on=index_fields,
                      suffixes=('_1', '_2'))

    # Keep track of which reads have ends flipped relative to clusters
    paired['flip'] = paired['cluster_end_1'] != 0
    paired = paired.drop(['cluster_end_1', 'cluster_end_2'], axis=1)
    paired = paired.reset_index()

    # Merge split read information into clusters
    data = split.merge(paired, on=['library_id', 'read_id', 'align_id_1', 'align_id_2'])
    data['position_1'] = data['position_1'].astype(int)
    data['position_2'] = data['position_2'].astype(int)

    # Check for empty split alignments
    if len(data.index) == 0:
      return pd.DataFrame()

    # Flip columns to make them consistent across clusters
    destruct.utils.misc.column_flip(data, data['flip'], 'chromosome_1', 'chromosome_2')
    destruct.utils.misc.column_flip(data, data['flip'], 'strand_1', 'strand_2')
    destruct.utils.misc.column_flip(data, data['flip'], 'position_1', 'position_2')

    # Track which end of the cluster was the seed end for this read
    data['seed_end'] = np.where(data['flip'], 1-data['read_end'], data['read_end'])

    def revcomp_inserted(row):
        if row['seed_end'] == 0:
            return row['inserted']
        else:
            return destruct.utils.misc.reverse_complement(row['inserted'])

    data['inserted'] = data.apply(revcomp_inserted, axis=1)

    data['inslen'] = data['inserted'].apply(len)

    data.set_index(['cluster_id', 'position_1', 'position_2', 'homology', 'inslen'], inplace=True)

    def calculate_consensus(dat):
        inserted = np.array([np.array(list(a)) for a in dat.values])
        consensus = list()
        for nt_list in inserted.T:
            consensus.append(collections.Counter(nt_list).most_common(1)[0][0])
        consensus = ''.join(consensus)
        return consensus

    agg_f = {'score':sum, 'read_id':len,
             'chromosome_1':max, 'chromosome_2':max,
             'strand_1':max, 'strand_2':max,
             'inserted':calculate_consensus}
    split_data = data.groupby(level=[0, 1, 2, 3, 4])\
                     .agg(agg_f)\
                     .rename(columns={'read_id':'count'})\
                     .reset_index()

    split_data = split_data.sort_values(['cluster_id', 'score'])

    split_data['breakpoint_id'] = split_data.groupby('cluster_id').cumcount(ascending=False)

    split_data = split_data[split_data['breakpoint_id'] < max_predictions_per_cluster]

    split_data['breakpoint_id'] += 1

    return split_data


def predict_breaks(clusters_filename, spanning_filename, split_filename, breakpoints_filename):

    clusters = pd.read_csv(clusters_filename, sep='\t', names=cluster_fields)

    if len(clusters.index) == 0:
        with open(breakpoints_filename, 'w'):
            pass
        return

    merge_columns = ['library_id', 'read_id', 'read_end', 'align_id']

    clusters_alignments = clusters[merge_columns].drop_duplicates()

    split_iter = pd.read_csv(split_filename, sep='\t', names=split_fields,
                             converters={'chromosome_1':str, 'chromosome_2':str, 'inserted':str},
                             iterator=True, chunksize=1000000)

    split_merge_columns = {'1':['library_id', 'read_id', 'read_end', 'align_id_1'],
                           '2':['library_id', 'read_id', 'read_end', 'align_id_2']}

    def filter_split(df):
        filtered = list()
        for side, left_merge_columns in split_merge_columns.items():
            df_2 = pd.merge(df, clusters_alignments,
                            left_on=left_merge_columns,
                            right_on=merge_columns,
                            how='inner')
            df_2 = df_2.drop(['align_id'], axis=1)
            filtered.append(df_2)
        return pd.concat(filtered).drop_duplicates()

    split = pd.concat([filter_split(chunk) for chunk in split_iter])

    split.loc[split['inserted'] == '.', 'inserted'] = ''

    spanning_iter = pd.read_csv(spanning_filename, sep='\t', names=spanning_fields,
                                converters={'chromosome':str},
                                iterator=True, chunksize=1000000)

    def filter_spanning(df):
        return pd.merge(df, clusters_alignments, how='inner')

    spanning = pd.concat([filter_spanning(chunk) for chunk in spanning_iter])

    predictions_0 = predict_breaks_spanning(clusters, spanning)

    predictions_1 = predict_breaks_split(clusters, split)

    predictions = pd.concat([predictions_0, predictions_1], ignore_index=True)
    predictions.sort_values('cluster_id', inplace=True)

    mate_scores = calculate_mate_score(clusters, spanning)
    predictions = predictions.merge(mate_scores, on='cluster_id')

    predictions.loc[predictions['inserted'] == '', 'inserted'] = '.'

    predictions = predictions[breakpoint_fields]
    predictions.to_csv(breakpoints_filename, sep='\t', index=False, header=False)


def calculate_cluster_weights(breakpoints_filename, weights_filename):

    epsilon = 0.0001
    itx_distance = 1000000000

    breakpoints = pd.read_csv(breakpoints_filename, sep='\t', names=breakpoint_fields,
                              converters={'chromosome_1':str, 'chromosome_2':str, 'inserted':str})

    breakpoints = breakpoints[breakpoints['breakpoint_id'] == 0]

    breakpoints['distance'] = np.absolute(breakpoints['position_1'] - breakpoints['position_2']) + 1.0
    breakpoints.loc[breakpoints['chromosome_1'] != breakpoints['chromosome_2'], 'distance'] = itx_distance
    breakpoints['weight'] = 1.0 + epsilon * np.log(breakpoints['distance'])

    breakpoints = breakpoints.sort_values('cluster_id')
    breakpoints[['cluster_id', 'weight']].to_csv(weights_filename, sep='\t', index=False, header=False)


def calculate_realignment_likelihoods(breakpoints_filename, realignments_filename, score_stats_filename,
                                      likelihoods_filename, match_score, fragment_mean, fragment_stddev):

    match_score = float(match_score)
    fragment_mean = float(fragment_mean)
    fragment_stddev = float(fragment_stddev)

    score_stats = pd.read_csv(score_stats_filename, sep='\t', names=score_stats_fields)

    breakpoints = pd.read_csv(breakpoints_filename, sep='\t', names=breakpoint_fields,
                              converters={'chromosome_1':str, 'chromosome_2':str, 'inserted':str})

    breakpoints.loc[breakpoints['inserted'] == '.', 'inserted'] = ''

    breakpoints['inslen'] = breakpoints['inserted'].apply(len)

    data = pd.read_csv(realignments_filename, sep='\t', names=realignment_fields)

    if len(data.index) == 0:
        with open(likelihoods_filename, 'w'):
            pass
        return

    data = data.merge(score_stats, on='aligned_length')

    # Alignment score likelihood and CDF
    data['max_score'] = match_score * data['aligned_length']
    data['score_diff'] = data['max_score'] - data['score']
    data['score_log_likelihood'] = np.log(data['expon_lda']) - \
                                          data['expon_lda'] * data['score_diff']
    data['score_log_cdf'] = -data['expon_lda'] * data['score_diff']

    # Unstack on cluster end
    index_fields = ['cluster_id', 'breakpoint_id', 'library_id', 'read_id']
    data_fields = ['read_end', 'aligned_length', 'template_length',
                   'score', 'score_log_likelihood', 'score_log_cdf']
    data = pd.merge(data.loc[data['cluster_end'] == 0, index_fields + data_fields].drop_duplicates(),
                    data.loc[data['cluster_end'] == 1, index_fields + data_fields].drop_duplicates(),
                    on=index_fields,
                    suffixes=('_1', '_2'))

    # For foldbacks and similar, a read could conceivable align to both sides of the breakpoint
    # remove situations in which one read end is aligned to both cluster ends
    data = data[data['read_end_1'] != data['read_end_2']]

    # Merge insert length from breakpoint predictions
    data = data.merge(breakpoints[['cluster_id', 'breakpoint_id', 'inslen']],
                      on=['cluster_id', 'breakpoint_id'])

    data['template_length'] = data['template_length_1'] + data['template_length_2'] + data['inslen']

    # Template length likelihood and CDF
    constant = 1. / ((2 * np.pi)**0.5 * fragment_stddev)
    data['length_z_score'] = (data['template_length'] - fragment_mean) / fragment_stddev
    data['length_log_likelihood'] = -np.log(constant) - np.square(data['length_z_score']) / 2.
    data['length_log_cdf'] = np.log(2. * scipy.stats.norm.sf(data['length_z_score'].abs()))

    data['log_likelihood'] = data['score_log_likelihood_1'] + \
                             data['score_log_likelihood_2'] + \
                             data['length_log_likelihood']
    data['log_cdf'] = data['score_log_cdf_1'] + \
                      data['score_log_cdf_2'] + \
                      data['length_log_cdf']

    index_fields = ['cluster_id', 'breakpoint_id', 'library_id', 'read_id']
    data = data.sort_values(index_fields + ['log_likelihood'])\
               .groupby(index_fields)\
               .last()\
               .reset_index()

    data = data[likelihoods_fields]

    data.to_csv(likelihoods_filename, sep='\t', index=False, header=False)


def select_clusters(clusters_filename,
                    breakpoints_filename, selected_breakpoints_filename,
                    likelihoods_filename, selected_likelihoods_filename):

    clusters = pd.read_csv(clusters_filename, sep='\t', names=cluster_fields,
                           usecols=['cluster_id', 'library_id', 'read_id'])
    clusters = clusters.drop_duplicates()

    breakpoints_iter = pd.read_csv(breakpoints_filename, sep='\t', names=breakpoint_fields,
                                   converters={'chromosome_1':str, 'chromosome_2':str, 'inserted':str},
                                   iterator=True, chunksize=1000000)

    cluster_ids = clusters[['cluster_id']].drop_duplicates()

    destruct.utils.streaming.read_select_write(breakpoints_iter, cluster_ids, selected_breakpoints_filename)

    likelihoods_iter = pd.read_csv(likelihoods_filename, sep='\t', names=likelihoods_fields,
                                   iterator=True, chunksize=1000000)

    destruct.utils.streaming.read_select_write(likelihoods_iter, clusters, selected_likelihoods_filename)


def select_breakpoint_prediction(likelihoods, template_length_min_threshold):
    """ Select and filter breakpoint predictions.

    Args:
        likelihoods(pandas.DataFrame): likelihoods table
        template_length_min_threshold (int): min template length filter

    Select the maximum likelihoods breakpoint prediction, preferring solutions
    with split reads.  Filter predictions based on minimum of template lengths on
    either side of the breakpoint.

    """

    # Calculate total breakpoint likelihood, max template lengths
    agg_f = {
        'log_likelihood':sum,
        'template_length_1':max,
        'template_length_2':max,
    }
    data = likelihoods.groupby(['cluster_id', 'breakpoint_id']).agg(agg_f).reset_index()

    # Select highest likelihood breakpoint predictions
    # Prefer a higher breakpoint id, thus preferring solutions with split reads
    data.sort_values(['cluster_id', 'log_likelihood', 'breakpoint_id'],
        ascending=[True, False, False], inplace=True)
    selected = data.groupby('cluster_id', sort=False).first().reset_index()

    # Filter by template length
    selected['template_length_min'] = selected[['template_length_1', 'template_length_2']].min(axis=1)
    selected.drop(['template_length_1', 'template_length_2'], axis=1, inplace=True)
    selected = selected[selected['template_length_min'] >= template_length_min_threshold]

    return selected


def select_predictions(breakpoints_filename, selected_breakpoints_filename,
                       likelihoods_filename, selected_likelihoods_filename,
                       mate_score_threshold, template_length_min_threshold,
                       min_alignment_log_likelihood):

    read_likelihoods_iter = pd.read_csv(likelihoods_filename, sep='\t', names=likelihoods_fields,
        usecols=['cluster_id', 'breakpoint_id', 'log_likelihood', 'template_length_1', 'template_length_2'],
        chunksize=int(1e7))

    read_likelihoods_iter = destruct.utils.streaming.group_aware_iter(read_likelihoods_iter, ['cluster_id'])

    selected = pd.concat([
        select_breakpoint_prediction(df, template_length_min_threshold)
        for df in read_likelihoods_iter], ignore_index=True)

    mate_score = pd.read_csv(breakpoints_filename, sep='\t', names=breakpoint_fields,
        converters={'chromosome_1':str, 'chromosome_2':str, 'inserted':str},
        usecols=['cluster_id', 'breakpoint_id', 'mate_score'])

    mate_score = mate_score.loc[mate_score['mate_score'] <= mate_score_threshold, ['cluster_id', 'breakpoint_id']]
    selected = selected.merge(mate_score)

    selected = selected[['cluster_id', 'breakpoint_id']].drop_duplicates()

    breakpoints_iter = pd.read_csv(breakpoints_filename, sep='\t', names=breakpoint_fields,
                                   converters={'chromosome_1':str, 'chromosome_2':str, 'inserted':str},
                                   iterator=True, chunksize=1000000)

    destruct.utils.streaming.read_select_write(breakpoints_iter, selected, selected_breakpoints_filename)

    likelihoods_iter = pd.read_csv(likelihoods_filename, sep='\t', names=likelihoods_fields,
                                   iterator=True, chunksize=1000000)

    def likelihoods_filter(chunk):
        chunk = chunk.merge(selected, how='inner')
        chunk = chunk[chunk['log_likelihood'] >= min_alignment_log_likelihood]
        return chunk

    destruct.utils.streaming.read_filter_write(likelihoods_iter, likelihoods_filter, selected_likelihoods_filename)





import sys
import tarfile
import numpy as np
import pandas as pd

import destruct.utils.plots


def load_align_data_hist(filename):

    data = pd.read_table(filename, sep='\t', names=['aligned_length', 'score'])

    align_data_hist = data.groupby(['aligned_length', 'score']).size()
    align_data_hist.name = 'freq'

    return align_data_hist.reset_index()


def create_score_stats(true_scores_filename, match_score, score_stats_filename):
    ''' Infer distribution of null alignment scores and true alignment scores from samples of null and true scores.
    The null samples may include some true scores so run a quick EM mixture model on the null samples to identify
    the actual distribution of null scores.  Output distributions for each alignment length.
    '''
    all_true_scores = load_align_data_hist(true_scores_filename)

    all_true_scores['penalty'] = match_score * all_true_scores['aligned_length'] - all_true_scores['score']

    score_stats = list()

    for aligned_length in all_true_scores['aligned_length'].unique():

        true_scores = all_true_scores.loc[all_true_scores['aligned_length'] == aligned_length]

        expon_lda = 1.0 / np.average(true_scores['penalty'], weights=true_scores['freq'])

        expon_lda = min(expon_lda, 1.0)

        score_stats.append((aligned_length, expon_lda))

    score_stats = pd.DataFrame(score_stats, columns=['aligned_length', 'expon_lda'])

    score_stats.to_csv(score_stats_filename, sep='\t', header=False, index=False)




import os
import subprocess
import shutil
import collections
import bisect
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import seaborn

import pypeliner

import destruct.benchmark.wrappers
import destruct.utils.download
import destruct.utils.misc
import destruct.utils.seq


class BreakpointDatabase(object):
    def __init__(self, breakpoints, id_col='breakpoint_id'):
        self.positions = collections.defaultdict(list)
        self.break_ids = collections.defaultdict(set)
        for idx, row in breakpoints.iterrows():
            for side in ('1', '2'):
                self.positions[(row['chromosome_'+side], row['strand_'+side])].append(row['position_'+side])
                self.break_ids[(row['chromosome_'+side], row['strand_'+side], row['position_'+side])].add((row[id_col], side))
        for key in self.positions.keys():
            self.positions[key] = sorted(self.positions[key])
    def query(self, row, extend=0):
        matched_ids = list()
        for side in ('1', '2'):
            chrom_strand_positions = self.positions[(row['chromosome_'+side], row['strand_'+side])]
            idx = bisect.bisect_left(chrom_strand_positions, row['position_'+side] - extend)
            side_matched_ids = list()
            while idx < len(chrom_strand_positions):
                pos = chrom_strand_positions[idx]
                dist = abs(pos - row['position_'+side])
                if pos >= row['position_'+side] - extend and pos <= row['position_'+side] + extend:
                    for break_id in self.break_ids[(row['chromosome_'+side], row['strand_'+side], pos)]:
                        side_matched_ids.append((break_id, dist))
                if pos > row['position_'+side] + extend:
                    break
                idx += 1
            matched_ids.append(side_matched_ids)
        matched_ids_bypos = list()
        for matched_id_1, dist_1 in matched_ids[0]:
            for matched_id_2, dist_2 in matched_ids[1]:
                if matched_id_1[0] == matched_id_2[0] and matched_id_1[1] != matched_id_2[1]:
                    matched_ids_bypos.append((dist_1 + dist_2, matched_id_1[0]))
        if len(matched_ids_bypos) == 0:
            return None
        return sorted(matched_ids_bypos)[0][1]


def create_tool_workflow(tool_info, bam_filenames, output_filename, raw_data_dir, **kwargs):
    workflow_module = __import__(tool_info['workflow']['module'], fromlist=[''])
    workflow_function = getattr(workflow_module, tool_info['workflow']['run_function'])

    if 'kwargs' in tool_info:
        kwargs.update(tool_info['kwargs'])

    try:
        os.makedirs(raw_data_dir)
    except OSError:
        pass

    return workflow_function(bam_filenames, output_filename, raw_data_dir, **kwargs)


def run_setup_function(tool_info, test_config, **kwargs):
    workflow_module = __import__(tool_info['workflow']['module'], fromlist=[''])
    setup_function = getattr(workflow_module, tool_info['workflow']['setup_function'])

    if 'kwargs' in tool_info:
        kwargs.update(tool_info['kwargs'])

    setup_function(test_config, **kwargs)


def create_roc_plot(sim_info, tool_defs, genome_fasta, simulated_filename, predicted_filename, annotated_filename, identified_filename, plot_filename):

    with PdfPages(plot_filename) as pdf:

        breakpoints = pd.read_csv(
            simulated_filename, sep='\t',
            converters={'breakpoint_id':str, 'chromosome_1':str, 'chromosome_2':str})
        breakpoints['inserted'] = breakpoints['inserted'].fillna('')

        breakpoints_db = BreakpointDatabase(breakpoints)

        results = pd.read_csv(
            predicted_filename, sep='\t',
            converters={'prediction_id':str, 'chromosome_1':str, 'chromosome_2':str})

        if 'inserted' not in results:
            results['inserted'] = ''
        results['inserted'] = results['inserted'].fillna('')

        def normalize_breakpoint(row, genome):
            norm_pos_1, norm_pos_2, homology = destruct.utils.misc.normalize_breakpoint(
                row['chromosome_1'], row['strand_1'], row['position_1'],
                row['chromosome_2'], row['strand_2'], row['position_2'], genome)

            row['normalized_position_1'] = norm_pos_1
            row['normalized_position_2'] = norm_pos_2
            row['homology'] = homology

            return row

        genome = dict(destruct.utils.seq.read_sequences(open(genome_fasta, 'r')))
        results = results.apply(normalize_breakpoint, axis=1, args=(genome,))
        del genome

        min_dist = 200
        def identify_true_positive(row):
            return breakpoints_db.query(row, min_dist)

        results['true_pos_id'] = results.apply(identify_true_positive, axis=1)

        results = results.merge(
            breakpoints,
            left_on='true_pos_id', right_on='breakpoint_id',
            suffixes=('', '_true'), how='left')

        def identify_exact_match(row):
            if ((row['chromosome_1'] == row['chromosome_1_true']) &
                (row['strand_1'] == row['strand_1_true']) &
                (row['normalized_position_1'] == row['position_1_true']) &
                (row['chromosome_2'] == row['chromosome_2_true']) &
                (row['strand_2'] == row['strand_2_true']) &
                (row['normalized_position_2'] == row['position_2_true'])):
                return True
            if ((row['chromosome_1'] == row['chromosome_2_true']) &
                (row['strand_1'] == row['strand_2_true']) &
                (row['normalized_position_1'] == row['position_2_true']) &
                (row['chromosome_2'] == row['chromosome_1_true']) &
                (row['strand_2'] == row['strand_1_true']) &
                (row['normalized_position_2'] == row['position_1_true'])):
                return True
            return False

        results['is_breakpoint_correct'] = results.apply(identify_exact_match, axis=1)
        results['is_homology_correct'] = (results['homology'] == results['homology_true'])
        results['is_inserted_correct'] = (results['inserted'] == results['inserted_true'])

        plot_cols = ['is_breakpoint_correct', 'is_homology_correct', 'is_inserted_correct']

        # Select the most correct breakpoint among multiple matching predictions
        data = (
            results.sort_values(['is_breakpoint_correct', 'is_inserted_correct'])
            .groupby('true_pos_id').first().sort_index().reset_index())

        data = data.set_index('true_pos_id')[plot_cols]
        data = data.reindex(breakpoints['breakpoint_id'].values)
        data = data.replace(True, 'correct').replace(False, 'incorrect').fillna('missing')
        data.columns = 'breakpoint', 'homology', 'inserted'
        data = data.stack()
        data.name = 'value'
        data.index.names = 'breakpoint_id', 'check'
        data = data.reset_index()

        g = seaborn.factorplot(col='check', x='value', kind='count', data=data)
        pdf.savefig(g.fig)

        results.to_csv(annotated_filename, sep='\t', index=False, na_rep='NA')

        features = tool_defs['features']
        invalid_value = -1e9

        identified = breakpoints[['breakpoint_id']]
        identified = identified.merge(results[['prediction_id', 'true_pos_id'] + features], left_on='breakpoint_id', right_on='true_pos_id', how='outer')
        identified = identified.drop('true_pos_id', axis=1)

        identified.to_csv(identified_filename, sep='\t', index=False, na_rep='NA')

        num_positive = int(sim_info['num_breakpoints'])

        y_test = results['true_pos_id'].notnull()

        tp_identified = np.sum(y_test)

        num_missed = num_positive - np.sum(y_test)

        y_test = np.concatenate([y_test, np.array([True]*num_missed)])

        if y_test.sum() == len(y_test):
            y_test = np.concatenate([y_test, np.array([False])])

        fig = plt.figure(figsize=(16,16))

        for feature in features:
            
            values = results[feature].values
            values = np.concatenate([values, np.array([invalid_value] * (len(y_test) - len(values)))])
            
            tp_counts = list()
            fp_counts = list()
            
            for threshold in np.sort(np.unique(values)):
                
                tp_counts.append(np.sum(y_test[values > threshold]))
                fp_counts.append(np.sum(1 - y_test[values > threshold]))
            
            plt.plot(fp_counts, tp_counts, label=feature.replace('_', ' '))

        plt.title('Features for {0} identified true positives'.format(tp_identified))
        plt.xlabel('False Positive Count')
        plt.ylabel('True Positive Count')
        plt.legend(loc="lower right")

        pdf.savefig(fig)

        results['Status'] = np.where(results['true_pos_id'].notnull(), 'True', 'False')

        for feature in features:

            fig = plt.figure(figsize=(16,16))

            seaborn.violinplot(x='Status', y=feature, data=results,
                               names=['True', 'False'])

            plt.title('True vs false for ' + feature)

            pdf.savefig(fig)


def create_genome(chromosomes, include_nonchromosomal, genome_fasta):
    """ Download and index a genome
    """

    destruct.utils.download.download_genome_fasta(genome_fasta,
                                                  chromosomes,
                                                  include_nonchromosomal)

    pypeliner.commandline.execute('bwa', 'index', genome_fasta)
    pypeliner.commandline.execute('samtools', 'faidx', genome_fasta)

    for extension in ('.fai', '.amb', '.ann', '.bwt', '.pac', '.sa'):
        os.rename(genome_fasta+extension, genome_fasta[:-4]+extension)


def create_ref_bam(in_ref_filename, in_bam_filename, out_ref_filename, out_bam_filename, chromosomes):
    """ Subset reference genome and bam by chromosomes
    """

    # Subset the chromosomes in the input reference genome
    samtools_faidx_cmd = [
        'samtools',
        'faidx',
        in_ref_filename,
    ]
    samtools_faidx_cmd += chromosomes
    samtools_faidx_cmd += ['>', out_ref_filename]
    pypeliner.commandline.execute(*samtools_faidx_cmd)

    # Fasta index of output reference
    out_ref_fai_filename = out_ref_filename + '.fai'
    pypeliner.commandline.execute('samtools', 'faidx', out_ref_filename)

    # Remove .tmp suffix if necessary
    if out_ref_filename.endswith('.tmp'):
        os.rename(out_ref_fai_filename, out_ref_filename[:-len('.tmp')] + '.fai')
        out_ref_fai_filename = out_ref_filename[:-len('.tmp')] + '.fai'

    # Create a new header
    original_header_filename = out_bam_filename + '.original_header.sam'
    new_header_filename = out_bam_filename + '.new_header.sam'
    pypeliner.commandline.execute('samtools', 'view', '-H', in_bam_filename, '>', original_header_filename)
    with open(original_header_filename, 'r') as in_f, open(new_header_filename, 'w') as out_f:
        for line in in_f:
            if line.startswith('@SQ'):
                keep = False
                for field in line.rstrip().split('\t'):
                    if field.startswith('SN:'):
                        chromosome = field[len('SN:'):]
                        if chromosome in chromosomes:
                            keep = True
                        break
                if keep:
                    out_f.write(line)
            else:
                out_f.write(line)

    # Create the chromosome subsetted bam file
    p = subprocess.Popen(['samtools', 'view', '-', '-b', '-t', out_ref_fai_filename, '-o', out_bam_filename], stdin=subprocess.PIPE)

    with open(new_header_filename, 'r') as f:
        shutil.copyfileobj(f, p.stdin)

    p2 = subprocess.Popen(['samtools', 'view', in_bam_filename] + chromosomes, stdout=p.stdin)

    if p2.wait() != 0:
        raise Exception('samtools view returned {}'.format(p2.returncode))

    p.stdin.close()
    if p.wait() != 0:
        raise Exception('samtools view returned {}'.format(p.returncode))

    # Samtools index the new bam
    pypeliner.commandline.execute('samtools', 'index', out_bam_filename)

    # Remove .tmp suffix if necessary
    if out_bam_filename.endswith('.tmp'):
        os.rename(out_bam_filename + '.bai', out_bam_filename[:-len('.tmp')] + '.bai')

    # Index for bwa alignment
    pypeliner.commandline.execute('bwa', 'index', out_ref_filename)

    # Remove .tmp suffix if necessary
    if out_ref_filename.endswith('.tmp'):
       for ext in ('.amb', '.ann', '.bwt', '.pac', '.sa'):
           os.rename(out_ref_filename + ext, out_ref_filename[:-len('.tmp')] + ext)


def partition_bam(original_filename, output_a_filename, output_b_filename, fraction_a):

    pypeliner.commandline.execute('destruct_bampartition',
                                  '-i', original_filename,
                                  '-a', output_a_filename,
                                  '-b', output_b_filename,
                                  '-f', fraction_a)

    pypeliner.commandline.execute('samtools', 'index', output_a_filename, output_a_filename[:-4]+'.bai')
    pypeliner.commandline.execute('samtools', 'index', output_b_filename, output_b_filename[:-4]+'.bai')


def samtools_sort_index(input_filename, output_filename):

    pypeliner.commandline.execute('samtools', 'sort', input_filename, '-o', output_filename)

    assert output_filename.endswith('.tmp')
    index_filename = output_filename[:-4] + '.bai'

    pypeliner.commandline.execute('samtools', 'index', output_filename)
    shutil.move(output_filename + '.bai', index_filename)


def samtools_merge_sort_index(output_filename, *input_filenames):

    sorted_filenames = list()

    for input_filename in input_filenames:
        pypeliner.commandline.execute('samtools', 'sort', input_filename, '-o', input_filename+'.sorted.bam')
        sorted_filenames.append(input_filename+'.sorted.bam')

    pypeliner.commandline.execute('samtools', 'merge', '-c', output_filename, *sorted_filenames)

    for sorted_filename in sorted_filenames:
        os.remove(sorted_filename)

    assert output_filename.endswith('.tmp')
    index_filename = output_filename[:-4] + '.bai'

    pypeliner.commandline.execute('samtools', 'index', output_filename, index_filename)


def samtools_sample_reheader(output_filename, input_filename, sample_name):
    header_filename = input_filename + '.header.sam'
    pypeliner.commandline.execute('samtools', 'view', '-H', input_filename, '>', header_filename)

    reheader_filename = input_filename + '.reheader.sam'
    with open(header_filename, 'r') as f_in, open(reheader_filename, 'w') as f_out:
        for line in f_in:
            fields = line.rstrip().split('\t')
            if fields[0] == '@RG':
                replaced = False
                for idx in range(len(fields)):
                    if fields[idx].startswith('SM:'):
                        fields[idx] = 'SM:' + sample_name
                        replaced = True
                if not replaced:
                    fields.append('SM:' + sample_name)
                line = '\t'.join(fields) + '\n'
            f_out.write(line)

    pypeliner.commandline.execute('samtools', 'reheader', reheader_filename, input_filename, '>', output_filename)

    assert output_filename.endswith('.tmp')
    index_filename = output_filename[:-4] + '.bai'

    pypeliner.commandline.execute('samtools', 'index', output_filename, index_filename)



import argparse
import destruct.somatic


def filter_annotate_breakpoints(**args):
    destruct.somatic.filter_annotate_breakpoints(
        args['input_breakpoint_filename'],
        args['input_breakpoint_library_filename'],
        args['control_ids'],
        args['output_breakpoint_filename'],
        args['output_breakpoint_library_filename'],
        patient_library_filename=args['patient_libraries_filename'],
    )


def add_arguments(argparser):
    argparser.add_argument('input_breakpoint_filename',
                           help='Input breakpoints filename')

    argparser.add_argument('input_breakpoint_library_filename',
                           help='Input breakpoint library filename')

    argparser.add_argument('output_breakpoint_filename',
                           help='Output breakpoints filename')

    argparser.add_argument('output_breakpoint_library_filename',
                           help='Output breakpoint library filename')

    argparser.add_argument('--control_ids', nargs='+',
                           help='Control/normal library ids')

    argparser.add_argument('--patient_libraries_filename', required=False,
                           help='Mapping from patient_id to library as tsv')

    argparser.set_defaults(func=filter_annotate_breakpoints)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser()

    add_arguments(argparser)

    args = vars(argparser.parse_args())
    func = args.pop('func')
    func(**args)



import pypeliner
import pypeliner.managed as mgd

import destruct.workflow
import destruct.create_ref_data
import destruct.benchmark.wrappers.destruct.tasks


def create_destruct_wrapper_workflow(bam_filenames, output_filename, raw_data_dir, control_id=None, config=None, ref_data_dir=None):
    workflow = pypeliner.workflow.Workflow(default_ctx={'mem': 4})

    workflow.setobj(
        obj=mgd.OutputChunks('sample_id'),
        value=list(bam_filenames.keys()),
    )

    workflow.subworkflow(
        name='run_destruct',
        func=destruct.workflow.create_destruct_workflow,
        args=(
            mgd.InputFile('bam', 'sample_id', fnames=bam_filenames),
            mgd.TempOutputFile('breakpoint_table'),
            mgd.TempOutputFile('breakpoint_library_table'),
            mgd.TempOutputFile('breakpoint_read_table'),
            config,
            ref_data_dir,
        ),
        kwargs={
            'raw_data_dir': raw_data_dir,
        },
    )

    workflow.transform(
        name='post_process',
        func=destruct.benchmark.wrappers.destruct.tasks.destruct_postprocess,
        args=(
            mgd.TempInputFile('breakpoint_table'),
            mgd.TempInputFile('breakpoint_library_table'),
            mgd.OutputFile(output_filename),
        ),
        kwargs={
            'control_id': control_id,
        }
    )

    return workflow


def setup_destruct(test_config, **kwargs):
    config = kwargs.get('config', {})
    
    if test_config.get('chromosomes', None) is not None:
        config = {
            'chromosomes': test_config['chromosomes'],
            'ensembl_assemblies': ['chromosome.{}'.format(c) for c in test_config['chromosomes']],
        }

    destruct.create_ref_data.create_ref_data(config, kwargs['ref_data_dir'])




import os

import pypeliner
import pypeliner.workflow
import pypeliner.managed as mgd

import destruct.defaultconfig


# Pypeliner contexts
locally = {'local': True}
lowmem = {'mem': 4, 'num_retry': 2, 'mem_retry_factor': 2}
medmem = {'mem': 8, 'num_retry': 2, 'mem_retry_factor': 2}
himem = {'mem': 16, 'num_retry': 2, 'mem_retry_factor': 2}


def create_destruct_workflow(
    bam_filenames,
    breakpoint_table,
    breakpoint_library_table,
    breakpoint_read_table,
    config,
    ref_data_dir,
    raw_data_dir=None,
):
    # Optionally cache raw reads for quicker rerun
    if raw_data_dir is not None:
        mgd_stats = mgd.File(os.path.join(raw_data_dir, '{bylibrary}_stats.txt'), 'bylibrary')
        mgd_reads_1 = mgd.File(os.path.join(raw_data_dir, '{bylibrary}_reads1.fq.gz'), 'bylibrary')
        mgd_reads_2 = mgd.File(os.path.join(raw_data_dir, '{bylibrary}_reads2.fq.gz'), 'bylibrary')
        mgd_sample_1 = mgd.File(os.path.join(raw_data_dir, '{bylibrary}_sample1.fq.gz'), 'bylibrary')
        mgd_sample_2 = mgd.File(os.path.join(raw_data_dir, '{bylibrary}_sample2.fq.gz'), 'bylibrary')

    else:
        mgd_stats = mgd.TempFile('stats.txt', 'bylibrary')
        mgd_reads_1 = mgd.TempFile('reads1.fq.gz', 'bylibrary')
        mgd_reads_2 = mgd.TempFile('reads2.fq.gz', 'bylibrary')
        mgd_sample_1 = mgd.TempFile('sample1.fq.gz', 'bylibrary')
        mgd_sample_2 = mgd.TempFile('sample2.fq.gz', 'bylibrary')

    config = destruct.defaultconfig.get_config(ref_data_dir, config)

    workflow = pypeliner.workflow.Workflow()

    # Set the library ids

    workflow.transform(
        name='create_lib_ids',
        func='destruct.tasks.create_library_ids',
        ret=mgd.TempOutputObj('library_id', 'bylibrary'),
        args=(list(bam_filenames.keys()),)
    )

    # Retrieve discordant reads and stats from bam files

    workflow.commandline(
        name='bamdisc',
        axes=('bylibrary',),
        ctx={'io': 1, 'mem': 8},
        args=(
            'destruct_bamdiscordantfastq',
            '-r',
            '-c', config['bam_max_soft_clipped'],
            '-f', config['bam_max_fragment_length'],
            '-b', mgd.InputFile('bam', 'bylibrary', fnames=bam_filenames),
            '-s', mgd_stats.as_output(),
            '--fastq1', mgd_reads_1.as_output(),
            '--fastq2', mgd_reads_2.as_output(),
            '-t', mgd.TempSpace('bamdisc.tempspace', 'bylibrary'),
            '-n', config['num_read_samples'],
            '--sample1', mgd_sample_1.as_output(),
            '--sample2', mgd_sample_2.as_output(),
        ),
    )

    workflow.subworkflow(
        name='destruct_fastq',
        func='destruct.workflow.create_destruct_fastq_workflow',
        args=(
            mgd_reads_1.as_input(),
            mgd_reads_2.as_input(),
            mgd_sample_1.as_input(),
            mgd_sample_2.as_input(),
            mgd_stats.as_input(),
            mgd.OutputFile(breakpoint_table),
            mgd.OutputFile(breakpoint_library_table),
            mgd.OutputFile(breakpoint_read_table),
            config,
            ref_data_dir,
        ),
        kwargs={
            'raw_data_dir': raw_data_dir,
        },
    )

    return workflow


def create_destruct_fastq_workflow(
    fastq1_filenames,
    fastq2_filenames,
    sample1_filenames,
    sample2_filenames,
    stats_filenames,
    breakpoint_table,
    breakpoint_library_table,
    breakpoint_read_table,
    config,
    ref_data_dir,
    raw_data_dir=None,
):
    workflow = pypeliner.workflow.Workflow()

    # Set the library ids

    workflow.transform(
        name='create_lib_ids',
        func='destruct.tasks.create_library_ids',
        ret=mgd.TempOutputObj('library_id', 'bylibrary'),
        args=(list(fastq1_filenames.keys()),)
    )

    workflow.transform(
        name='readstats',
        axes=('bylibrary',),
        ctx=lowmem,
        func='destruct.tasks.read_stats',
        ret=mgd.TempOutputObj('stats', 'bylibrary'),
        args=(
            mgd.InputFile('stats.txt', 'bylibrary', fnames=stats_filenames),
            config['fragment_length_num_stddevs'],
        ),
    )

    # Align a sample of reads and calculate alignment statistics

    workflow.transform(
        name='prepseed_sample',
        axes=('bylibrary',),
        ctx=medmem,
        func='destruct.tasks.prepare_seed_fastq',
        args=(
            mgd.InputFile('sample1.fq.gz', 'bylibrary', fnames=sample1_filenames),
            mgd.InputFile('sample2.fq.gz', 'bylibrary', fnames=sample2_filenames),
            36,
            mgd.TempOutputFile('sample.seed', 'bylibrary'),
        ),
    )

    workflow.commandline(
        name='bwtrealign_sample',
        axes=('bylibrary',),
        ctx=medmem,
        args=(
            'bowtie',
            config['genome_fasta'],
            mgd.TempInputFile('sample.seed', 'bylibrary'),
            '--chunkmbs', '512',
            '-k', '1000',
            '-m', '1000',
            '--strata',
            '--best',
            '-S',
            '|',
            'destruct_aligntrue',
            '-a', '-',
            '-1', mgd.InputFile('sample1.fq.gz', 'bylibrary', fnames=sample1_filenames),
            '-2', mgd.InputFile('sample2.fq.gz', 'bylibrary', fnames=sample2_filenames),
            '-r', config['genome_fasta'],
            '-g', config['gap_score'],
            '-x', config['mismatch_score'],
            '-m', config['match_score'],
            '--flmin', mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_min'),
            '--flmax', mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_max'),
            '-s', mgd.TempOutputFile('samples.align.true', 'bylibrary'),
        ),
    )

    workflow.transform(
        name='scorestats',
        axes=('bylibrary',),
        ctx=medmem,
        func='destruct.score_stats.create_score_stats',
        args=(
            mgd.TempInputFile('samples.align.true', 'bylibrary'),
            config['match_score'],
            mgd.TempOutputFile('score.stats', 'bylibrary'),
        ),
    )

    # Split discordant fastqs and align

    workflow.transform(
        name='splitfastq1',
        axes=('bylibrary',),
        ctx=lowmem,
        func='destruct.tasks.split_fastq',
        args=(
            mgd.InputFile('reads1.fq.gz', 'bylibrary', fnames=fastq1_filenames),
            int(config['reads_per_split']),
            mgd.TempOutputFile('reads1', 'bylibrary', 'byread'),
        ),
    )

    workflow.transform(
        name='splitfastq2',
        axes=('bylibrary',),
        ctx=lowmem,
        func='destruct.tasks.split_fastq',
        args=(
            mgd.InputFile('reads2.fq.gz', 'bylibrary', fnames=fastq2_filenames),
            int(config['reads_per_split']),
            mgd.TempOutputFile('reads2', 'bylibrary', 'byread', axes_origin=[]),
        ),
    )

    workflow.transform(
        name='prepseed',
        axes=('bylibrary', 'byread'),
        ctx=medmem,
        func='destruct.tasks.prepare_seed_fastq',
        args=(
            mgd.TempInputFile('reads1', 'bylibrary', 'byread'),
            mgd.TempInputFile('reads2', 'bylibrary', 'byread'),
            36,
            mgd.TempOutputFile('reads.seed', 'bylibrary', 'byread'),
        ),
    )

    workflow.commandline(
        name='bwtrealign',
        axes=('bylibrary', 'byread'),
        ctx=medmem,
        args=(
            'bowtie',
            config['genome_fasta'],
            mgd.TempInputFile('reads.seed', 'bylibrary', 'byread'),
            '--chunkmbs', '512',
            '-k', '1000',
            '-m', '1000',
            '--strata',
            '--best',
            '-S',
            '|',
            'destruct_realign2',
            '-l', mgd.TempInputObj('library_id', 'bylibrary'),
            '-a', '-',
            '-1', mgd.TempInputFile('reads1', 'bylibrary', 'byread'),
            '-2', mgd.TempInputFile('reads2', 'bylibrary', 'byread'),
            '-r', config['genome_fasta'],
            '-g', config['gap_score'],
            '-x', config['mismatch_score'],
            '-m', config['match_score'],
            '--flmin', mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_min'),
            '--flmax', mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_max'),
            '--tchimer', config['chimeric_threshold'],
            '--talign', config['alignment_threshold'],
            '--pchimer', config['chimeric_prior'],
            '--tvalid', config['readvalid_threshold'],
            '-z', mgd.TempInputFile('score.stats', 'bylibrary'),
            '--span', mgd.TempOutputFile('spanning.alignments', 'bylibrary', 'byread'),
            '--split', mgd.TempOutputFile('split.alignments', 'bylibrary', 'byread'),
        ),
    )

    workflow.transform(
        name='merge_spanning_1',
        axes=('bylibrary',),
        ctx=lowmem,
        func='destruct.tasks.merge_files_by_line',
        args=(
            mgd.TempInputFile('spanning.alignments', 'bylibrary', 'byread'),
            mgd.TempOutputFile('spanning.alignments_1', 'bylibrary'),
        ),
    )

    workflow.commandline(
        name='filterreads',
        axes=('bylibrary',),
        ctx=lowmem,
        args=(
            'destruct_filterreads',
            '-n', '2',
            '-a', mgd.TempInputFile('spanning.alignments_1', 'bylibrary'),
            '-r', config['satellite_regions'],
            '>', mgd.TempOutputFile('spanning.alignments', 'bylibrary'),
        ),
    )

    workflow.transform(
        name='merge_split_1',
        axes=('bylibrary',),
        ctx=lowmem,
        func='destruct.tasks.merge_files_by_line',
        args=(
            mgd.TempInputFile('split.alignments', 'bylibrary', 'byread'),
            mgd.TempOutputFile('split.alignments', 'bylibrary'),
        ),
    )

    workflow.transform(
        name='merge_spanning_2',
        ctx=lowmem,
        func='destruct.tasks.merge_alignment_files',
        args=(
            mgd.TempInputFile('spanning.alignments', 'bylibrary'),
            mgd.TempOutputFile('spanning.alignments'),
            mgd.TempInputObj('library_id', 'bylibrary'),
        ),
    )

    workflow.transform(
        name='merge_split_2',
        ctx=lowmem,
        func='destruct.tasks.merge_alignment_files',
        args=(
            mgd.TempInputFile('split.alignments', 'bylibrary'),
            mgd.TempOutputFile('split.alignments'),
            mgd.TempInputObj('library_id', 'bylibrary'),
        ),
    )

    # Cluster spanning reads

    workflow.transform(
        name='generate_chrom_args',
        func='destruct.tasks.generate_chromosome_args',
        ret=mgd.TempOutputObj('chrom.args', 'bychromarg'),
        args=(config['chromosomes'],)
    )

    workflow.transform(
        name='write_stats_table',
        ctx=lowmem,
        func='destruct.tasks.write_stats_table',
        args=(
            mgd.TempInputObj('library_id', 'bylibrary'),
            mgd.TempInputObj('stats', 'bylibrary'),
            mgd.TempOutputFile('libstats.tsv'),
        ),
    )

    workflow.commandline(
        name='cluster',
        axes=('bychromarg',),
        ctx=medmem,
        args=(
            'destruct_mclustermatepairs',
            '-a', mgd.TempInputFile('spanning.alignments'),
            '-s', mgd.TempInputFile('libstats.tsv'),
            '-c', mgd.TempOutputFile('clusters', 'bychromarg'),
            mgd.TempInputObj('chrom.args', 'bychromarg'),
            '--clustmin', config['cluster_readcount_threshold'],
            '--fragmax', config['fragment_length_max'],
        ),
    )
    
    # Predict breakpoints from split reads

    workflow.transform(
        name='predict_breaks',
        axes=('bychromarg',),
        ctx=medmem,
        func='destruct.predict_breaks.predict_breaks',
        args=(
            mgd.TempInputFile('clusters', 'bychromarg'),
            mgd.TempInputFile('spanning.alignments'),
            mgd.TempInputFile('split.alignments'),
            mgd.TempOutputFile('breakpoints_2', 'bychromarg'),
        ),
    )

    workflow.transform(
        name='merge_clusters',
        ctx=lowmem,
        func='destruct.tasks.merge_clusters',
        args=(
            mgd.TempInputFile('clusters', 'bychromarg'),
            mgd.TempInputFile('breakpoints_2', 'bychromarg'),
            mgd.TempOutputFile('clusters'),
            mgd.TempOutputFile('breakpoints_2'),
            mgd.TempOutputFile('merge_clusters.debug'),
        ),
    )

    # Realign reads to breakpoints

    workflow.commandline(
        name='realigntobreaks',
        axes=('bylibrary', 'byread'),
        ctx=medmem,
        args=(
            'destruct_realigntobreaks2',
            '-r', config['genome_fasta'],
            '-b', mgd.TempInputFile('breakpoints_2'),
            '-c', mgd.TempInputFile('clusters'),
            '-g', config['gap_score'],
            '-x', config['mismatch_score'],
            '-m', config['match_score'],
            '--flmax', mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_max'),
            '--span', mgd.TempInputFile('spanning.alignments', 'bylibrary', 'byread'),
            '-1', mgd.TempInputFile('reads1', 'bylibrary', 'byread'),
            '-2', mgd.TempInputFile('reads2', 'bylibrary', 'byread'),
            '--realignments', mgd.TempOutputFile('realignments', 'bylibrary', 'byread'),
        ),
    )

    # Calculate likelihoods based on realignments

    workflow.transform(
        name='calculate_realignment_likelihoods',
        axes=('bylibrary', 'byread'),
        ctx=medmem,
        func='destruct.predict_breaks.calculate_realignment_likelihoods',
        args=(
            mgd.TempInputFile('breakpoints_2'),
            mgd.TempInputFile('realignments', 'bylibrary', 'byread'),
            mgd.TempInputFile('score.stats', 'bylibrary'),
            mgd.TempOutputFile('likelihoods_2', 'bylibrary', 'byread'),
            config['match_score'],
            mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_mean'),
            mgd.TempInputObj('stats', 'bylibrary').prop('fragment_length_stddev'),
        ),
    )

    workflow.transform(
        name='merge_likelihoods_1',
        axes=('bylibrary',),
        ctx=lowmem,
        func='destruct.tasks.merge_sorted_files_by_line',
        args=(
            mgd.TempInputFile('likelihoods_2', 'bylibrary', 'byread'),
            mgd.TempOutputFile('likelihoods_2', 'bylibrary'),
            mgd.TempSpace('merge_likelihoods_1_temp', 'bylibrary'),
            '1',
        ),
    )

    workflow.transform(
        name='merge_likelihoods_2',
        ctx=lowmem,
        func='destruct.tasks.merge_sorted_files_by_line',
        args=(
            mgd.TempInputFile('likelihoods_2', 'bylibrary'),
            mgd.TempOutputFile('likelihoods_2'),
            mgd.TempSpace('merge_likelihoods_2_temp'),
            '1',
        ),
    )

    # Set cover for multi mapping reads

    workflow.transform(
        name='calc_weights',
        ctx=medmem,
        func='destruct.predict_breaks.calculate_cluster_weights',
        args=(
            mgd.TempInputFile('breakpoints_2'),
            mgd.TempOutputFile('cluster_weights'),
        ),
    )

    workflow.commandline(
        name='setcover',
        ctx=medmem,
        args=(
            'destruct_setcover',
            '-c', mgd.TempInputFile('clusters'),
            '-w', mgd.TempInputFile('cluster_weights'),
            '-a', mgd.TempOutputFile('clusters_setcover'),
        ),
    )

    # Select cluster based on setcover

    workflow.transform(
        name='select_clusters',
        ctx=medmem,
        func='destruct.predict_breaks.select_clusters',
        args=(
            mgd.TempInputFile('clusters_setcover'),
            mgd.TempInputFile('breakpoints_2'),
            mgd.TempOutputFile('breakpoints_1'),
            mgd.TempInputFile('likelihoods_2'),
            mgd.TempOutputFile('likelihoods_1'),
        ),
    )

    # Select prediction based on max likelihood

    workflow.transform(
        name='select_predictions',
        ctx=himem,
        func='destruct.predict_breaks.select_predictions',
        args=(
            mgd.TempInputFile('breakpoints_1'),
            mgd.TempOutputFile('breakpoints'),
            mgd.TempInputFile('likelihoods_1'),
            mgd.TempOutputFile('likelihoods'),
            config['mate_score_threshold'],
            config['template_length_min_threshold'],
            config['min_alignment_log_likelihood'],
        ),
    )

    # Optionally tabulate supporting reads

    workflow.transform(
        name='tabreads',
        ctx=medmem,
        func='destruct.tasks.tabulate_reads',
        args=(
            mgd.TempInputFile('clusters_setcover'),
            mgd.TempInputFile('likelihoods'),
            mgd.TempInputObj('library_id', 'bylibrary'),
            mgd.InputFile('reads1.fq.gz', 'bylibrary', fnames=fastq1_filenames),
            mgd.InputFile('reads2.fq.gz', 'bylibrary', fnames=fastq2_filenames),
            mgd.TempOutputFile('breakreads.table.unsorted'),
        ),
    )

    workflow.commandline(
        name='sortreads',
        ctx=medmem,
        args=(
            'sort', '-n',
            mgd.TempInputFile('breakreads.table.unsorted'),
            '>', mgd.OutputFile(breakpoint_read_table),
        ),
    )


    # Tabulate results

    workflow.transform(
        name='tabulate',
        ctx=himem,
        func='destruct.tasks.tabulate_results',
        args=(
            mgd.TempInputFile('breakpoints'),
            mgd.TempInputFile('likelihoods'),
            mgd.TempInputObj('library_id', 'bylibrary'),
            config['genome_fasta'],
            config['gtf_filename'],
            config['dgv_filename'],
            mgd.OutputFile(breakpoint_table),
            mgd.OutputFile(breakpoint_library_table),
        ),
    )

    return workflow




import argparse
import yaml

import destruct.benchmark.align.bwa.workflow
import destruct.benchmark.destruct_test
import destruct.benchmark.create_breakpoint_simulation


if __name__ == '__main__':

    argparser = argparse.ArgumentParser()

    argparser.add_argument('ref_data_dir',
                           help='Reference genomes directory')

    argparser.add_argument('tool_defs',
                           help='Tool Definition Filename')

    argparser.add_argument('--chromosomes', nargs='*', type=str, default=None,
                           help='Reference chromosomes')

    args = vars(argparser.parse_args())

    yaml_text = open(args['tool_defs']).read().format(ref_data_dir=args['ref_data_dir'])
    tool_defs = yaml.load(yaml_text)

    ctx = {'mem':4}

    test_config = {}
    if args.get('chromosomes', None) is not None:
        test_config['chromosomes'] = args['chromosomes']

    for tool_name, tool_info in tool_defs.items():
        destruct.benchmark.destruct_test.run_setup_function(tool_info, test_config)




import os
import shutil
import subprocess
import hashlib


def hash_kwargs(kwargs):
    kwargs = tuple(sorted(list(kwargs.items())))
    kwargs_hash = hashlib.sha224(repr(kwargs)).hexdigest()
    return kwargs_hash


class Sentinal(object):
    def __init__(self, sentinal_filename):
        self.sentinal_filename = sentinal_filename
    @property
    def unfinished(self):
        if os.path.exists(self.sentinal_filename):
            print ('sentinal file ' + self.sentinal_filename + ' exists')
            return False
        return True
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            with open(self.sentinal_filename, 'w'):
                pass


class SentinalFactory(object):
    def __init__(self, filename_prefix, kwargs):
        self.filename_prefix = filename_prefix
        self.kwargs_hash = hash_kwargs(kwargs)
    def __call__(self, name):
        return Sentinal(self.filename_prefix + name + '_' + self.kwargs_hash)


def makedirs(directory):
    try:
        os.makedirs(directory)
    except OSError as e:
        if e.errno != 17:
            raise


def rmtree(directory):
    try:
        shutil.rmtree(directory)
    except OSError as e:
        if e.errno != 2:
            raise


def remove(filename):
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != 2:
            raise


def symlink(filename, link_name=None, link_directory=None):
    if link_name is None:
        link_name = os.path.basename(filename)
    if link_directory is None:
        link_directory = os.getcwd()
    link_filename = os.path.join(link_directory, link_name)
    remove(link_filename)
    filename = os.path.abspath(filename)
    os.symlink(filename, link_filename)
    return link_filename


class CurrentDirectory(object):
    def __init__(self, directory):
        self.directory = directory
    def __enter__(self):
        self.prev_directory = os.getcwd()
        makedirs(self.directory)
        os.chdir(self.directory)
    def __exit__(self, *args):
        os.chdir(self.prev_directory)


def wget_file(url, filename):
    makedirs(os.path.dirname(filename))
    subprocess.check_call(['wget', '--no-check-certificate', url, '-O', filename])


def wget_file_gunzip(url, filename):
    makedirs(os.path.dirname(filename))
    subprocess.check_call(['wget', '--no-check-certificate', url, '-O', filename+'.gz'])
    remove(filename)
    subprocess.check_call(['gunzip', filename+'.gz'])


class SafeWriteFile(object):
    def __init__(self, filename):
        self.filename = filename
        self.temp_filename = filename + '.tmp'
    def __enter__(self):
        return self.temp_filename
    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is None:
            os.rename(self.temp_filename, self.filename)





import random
import os
import subprocess
import pandas as pd

import destruct.utils.seq
import destruct.utils.misc


def random_chromosome_position(genome, dist_to_end):
    """ Generate random chromosome, strand position.

     > Note that positions are 1-based
    """

    genome_length = sum([len(chr) for chr in genome.values()])
    while True:
        position = random.randint(1, genome_length + 1)
        for id, chr in genome.items():
            if position <= len(chr):
                if position <= dist_to_end or len(chr) - position < dist_to_end:
                    break
                return id, random.choice(['+', '-']), position
            position -= len(chr)


def retrieve_upstream_sequence(genome, chromosome, strand, position, length):
    """ Retrieve sequence upstream of a position, including that position
    """

    if strand == '+':
        end = position
        start = position - length + 1
    else:
        start = position
        end = position + length - 1

    # Positions assumed 1-based, closed region
    start -= 1

    return genome[chromosome][start:end]


def retrieve_downstream_sequence(genome, chromosome, strand, position, length):
    """ Retrieve sequence downstream of a position, not including that position
    """

    if strand == '+':
        start = position + 1
        end = start + length - 1
    else:
        end = position - 1
        start = end - length + 1

    # Positions assumed 1-based, closed region
    start -= 1

    return genome[chromosome][start:end]


def create_distinct(sequence1, sequence2, length):
    """ Create an insertion sequence distinct from reference genome sequence
    """

    sequence1 = sequence1.upper()
    sequence2 = sequence2.upper()
    distinct = ''
    for idx, nts in enumerate(zip(sequence1, sequence2)):
        while len(distinct) < length:
            nt = random.choice(['A', 'C', 'T', 'G'])
            if nt not in nts:
                distinct = distinct + nt
                break
    return distinct


def max_similar(seq1, seq2):
    """ Calculate the maximum number of identical nucleotides
    """

    similar = 0
    for nt1, nt2 in zip(seq1, seq2):
        if nt1 != nt2:
            break
        similar += 1
    return similar


def create_random_breakpoint(genome, adjacent_length, num_inserted, required_homology=None):
    """ Create a random breakpoint with the required parameters
    """

    while True:
        chr1, str1, pos1 = random_chromosome_position(genome, adjacent_length)
        chr2, str2, pos2 = random_chromosome_position(genome, adjacent_length)
        downstream1 = retrieve_downstream_sequence(genome, chr1, str1, pos1, adjacent_length)
        downstream2 = retrieve_downstream_sequence(genome, chr2, str2, pos2, adjacent_length)
        sequence1 = retrieve_upstream_sequence(genome, chr1, str1, pos1, adjacent_length)
        sequence2 = retrieve_upstream_sequence(genome, chr2, str2, pos2, adjacent_length)
        if str1 != '+':
            sequence1 = destruct.utils.misc.reverse_complement(sequence1)
            downstream1 = destruct.utils.misc.reverse_complement(downstream1)
        if str2 != '-':
            sequence2 = destruct.utils.misc.reverse_complement(sequence2)
            downstream2 = destruct.utils.misc.reverse_complement(downstream2)
        inserted = create_distinct(downstream1, downstream2, num_inserted)
        homology = max_similar(sequence1[::-1], downstream2[::-1]) + max_similar(sequence2, downstream1)
        if required_homology is not None and homology != required_homology:
            continue
        sequence = sequence1 + inserted + sequence2
        if 'N' in sequence or 'n' in sequence:
            continue
        if inserted == '':
            norm_pos1, norm_pos2, homology_check = destruct.utils.misc.normalize_breakpoint(chr1, str1, pos1, chr2, str2, pos2, genome)
            if homology_check != homology:
                raise Exception(
                    '''homology {} doesnt match expected homology {} for 
                    breakpoint {} {} {}, {} {} {}'''.format(
                        homology_check, homology,
                        chr1, str1, pos1, chr2, str2, pos2))
            pos1, pos2 = norm_pos1, norm_pos2
        return chr1, str1, pos1, chr2, str2, pos2, inserted, sequence, homology

def simulate(sim_info, read_count, sequences_fasta, reads1, reads2, random_reads=True):
    """ Simulate reads
    """

    temp_prefix = reads1 + '.dwgsim.temp.prefix'
    dwgsim_command = ['dwgsim']
    dwgsim_command += ['-z', sim_info['dwgsim_seed']]
    dwgsim_command += ['-d', sim_info['fragment_mean']]
    dwgsim_command += ['-s', sim_info['fragment_stddev']]
    dwgsim_command += ['-N', read_count]
    dwgsim_command += ['-1', sim_info['read_length']]
    dwgsim_command += ['-2', sim_info['read_length']]
    dwgsim_command += ['-H']
    if not random_reads:
        dwgsim_command += ['-y', '0']
    if sim_info.get('perfect_reads', '') == 'yes':
        dwgsim_command += ['-e', '0']
        dwgsim_command += ['-E', '0']
        dwgsim_command += ['-r', '0']
        dwgsim_command += ['-R', '0']
    dwgsim_command += [sequences_fasta]
    dwgsim_command += [temp_prefix]
    dwgsim_command = [str(a) for a in dwgsim_command]
    print (' '.join(dwgsim_command))
    dwgsim_retcode = subprocess.call(dwgsim_command)
    assert(dwgsim_retcode == 0)
    os.rename(temp_prefix + '.bwa.read1.fastq', reads1)
    os.rename(temp_prefix + '.bwa.read2.fastq', reads2)
    try:
        os.remove(temp_prefix + '.bfast.fastq')
    except OSError:
        pass
    try:
        os.remove(temp_prefix + '.mutations.txt')
    except OSError:
        pass
    try:
        os.remove(temp_prefix + '.mutations.vcf')
    except OSError:
        pass

def create_breakpoints(sim_info, genome_fasta, breakpoints_fasta, breakpoints_info):
    """ Create a set of simulated breakpoints
    """

    random.seed(int(sim_info['breakpoints_seed']))
    genome = dict(destruct.utils.seq.read_sequences(open(genome_fasta, 'r')))
    info_table = []
    with open(breakpoints_fasta, 'w') as fasta:
        for idx in range(int(sim_info['num_breakpoints'])):
            adjacent_length = int(sim_info['adjacent_length'])
            if sim_info['random_break_features']:
                homology = random.randint(0, int(sim_info['homology']))
                if homology == 0 and random.random() < 0.1:
                    num_inserted = random.randint(0, int(sim_info['num_inserted']))
                else:
                    num_inserted = 0
            else:
                num_inserted = int(sim_info['num_inserted'])
                homology = int(sim_info['homology'])
            chr1, str1, pos1, chr2, str2, pos2, inserted, sequence, homology = create_random_breakpoint(
                genome, adjacent_length, num_inserted, homology)
            fasta.write('>{0}\n{1}\n'.format(idx, sequence))
            info_table.append(
                {
                    'breakpoint_id': idx,
                    'chromosome_1': chr1,
                    'strand_1': str1,
                    'position_1': pos1,
                    'chromosome_2': chr2,
                    'strand_2': str2,
                    'position_2': pos2,
                    'inserted': inserted,
                    'homology': homology,
                }
            )
    info_table = pd.DataFrame(info_table)
    info_table.to_csv(breakpoints_info, sep='\t', index=False)

def create(sim_info, genome_fasta, breakpoints_fasta, breakpoints_info, concordant1, concordant2, discordant1, discordant2):
    """ Create simulated breakpoints and simulate reads from those breakpoints
    """

    create_breakpoints(sim_info, genome_fasta, breakpoints_fasta, breakpoints_info)
    sequences_size = sum([len(seq) for id, seq in destruct.utils.seq.read_sequences(open(breakpoints_fasta, 'r'))])
    read_count = int(float(sim_info['coverage']) * sequences_size / float(sim_info['fragment_mean']))
    simulate(sim_info, read_count, breakpoints_fasta, discordant1, discordant2, False)
    simulate(sim_info, sim_info['num_concordant'], genome_fasta, concordant1, concordant2)



import os
import pysam
import pandas as pd

import pypeliner.commandline


def run_lumpyexpress(bam_filenames, splitters_filenames, discordants_filenames, results_vcf):
    bam_arg = []
    splitters_arg = []
    discordants_arg = []
    for sample_id in bam_filenames.keys():
        bam_arg.append(bam_filenames[sample_id])
        splitters_arg.append(splitters_filenames[sample_id])
        discordants_arg.append(discordants_filenames[sample_id])

    bam_arg = ','.join(bam_arg)
    splitters_arg = ','.join(splitters_arg)
    discordants_arg = ','.join(discordants_arg)

    pypeliner.commandline.execute(
        'lumpyexpress',
        '-B', bam_arg,
        '-S', splitters_arg,
        '-D', discordants_arg,
        '-o', results_vcf,
    )


def vcf_to_bcf(vcf_filename, bcf_filename):
    sorted_vcf = vcf_filename + '.sorted'
    pypeliner.commandline.execute('vcf-sort', vcf_filename, '>', sorted_vcf)

    gz_vcf = vcf_filename + '.sorted.gz'
    pypeliner.commandline.execute('bgzip', sorted_vcf)

    gz_vcf_tbi = vcf_filename + '.sorted.gz.tbi'
    pypeliner.commandline.execute('tabix', gz_vcf)

    pypeliner.commandline.execute('bcftools', 'convert', '-O', 'b', gz_vcf, '-o', bcf_filename)

    os.remove(gz_vcf)
    os.remove(gz_vcf_tbi)


def convert_bcf(bcf_filename, output_filename, control_id=None):
    bcf_reader = pysam.VariantFile(bcf_filename, 'rb')

    breakpoint_table = list()
    breakpoint_library_table = list()

    for row in bcf_reader:
        if row.info.get('SECONDARY', False):
            continue

        prediction_id = row.id

        assert row.info['SVTYPE'] in ('DEL', 'DUP', 'INV', 'BND')

        chrom_1 = row.chrom
        coord_1 = row.pos
        if row.info['SVTYPE'] == 'BND':
            if '[' in row.alts[0]:
                chrom_2, coord_2 = row.alts[0].split('[')[1].split(':')
                coord_2 = int(coord_2)
            elif ']' in row.alts[0]:
                chrom_2, coord_2 = row.alts[0].split(']')[1].split(':')
                coord_2 = int(coord_2)
        else:
            chrom_2 = row.chrom
            coord_2 = row.stop
            assert coord_1 < coord_2

        breakpoint_ids = []
        assert isinstance(row.info['STRANDS'], tuple)
        for strand_info, breakpoint_id in zip(row.info['STRANDS'], 'AB'):
            (strand_1, strand_2), breakpoint_read_count = strand_info.split(':')

            if row.info['SVTYPE'] == 'DEL':
                assert (strand_1, strand_2) == ('+', '-')

            elif row.info['SVTYPE'] == 'DUP':
                assert (strand_1, strand_2) == ('-', '+')

            elif row.info['SVTYPE'] == 'DUP':
                assert strand_1 == strand_2

            svtype = row.info['SVTYPE']

            breakpoint_table.append((prediction_id + '_' + breakpoint_id, chrom_1, chrom_2, strand_1, strand_2, coord_1, coord_2, svtype, breakpoint_read_count))
            breakpoint_ids.append(prediction_id + '_' + breakpoint_id)

        for sample, call in row.samples.items():
            num_spanning = call['PE']
            num_split = call['SR']

            for breakpoint_id in breakpoint_ids:
                breakpoint_library_table.append((breakpoint_id, sample, num_spanning, num_split))

    breakpoint_table = pd.DataFrame(
        breakpoint_table,
        columns=[
            'prediction_id',
            'chromosome_1', 'chromosome_2',
            'strand_1', 'strand_2',
            'position_1', 'position_2',
            'svtype',
            'breakpoint_read_count',
        ]
    )

    breakpoint_library_table = pd.DataFrame(
        breakpoint_library_table,
        columns=[
            'prediction_id',
            'library',
            'num_spanning',
            'num_split'
        ]
    )

    spanning_read_counts = breakpoint_library_table.groupby('prediction_id')['num_spanning'].sum().reset_index()
    split_read_counts = breakpoint_library_table.groupby('prediction_id')['num_split'].sum().reset_index()

    total_read_counts = breakpoint_library_table.set_index(['prediction_id', 'library']).sum(axis=1).unstack()
    total_read_counts.columns = [a + '_count' for a in total_read_counts.columns]
    total_read_counts = total_read_counts.reset_index()

    breakpoint_table = breakpoint_table.merge(spanning_read_counts, on='prediction_id')
    breakpoint_table = breakpoint_table.merge(split_read_counts, on='prediction_id')
    breakpoint_table = breakpoint_table.merge(total_read_counts, on='prediction_id')

    # Filter based on evidence in control dataset
    if control_id is not None:
         breakpoint_table = breakpoint_table[breakpoint_table['{0}_count'.format(control_id)] == 0]          

    breakpoint_table.to_csv(output_filename, sep='\t', index=False)





import pandas as pd

def read_select_write(df_iter, select, out_filename):
    """ Write subset of streamed data.

    Args:
        df_iter (iter of pandas.DataFrame): streamed data
        select (pandas.DataFrame): subset for filtering
        out_filename (str): output tsv filename

    """

    with open(out_filename, 'w') as out_file:
        for chunk in df_iter:
            chunk = chunk.merge(select, how='inner')
            chunk.to_csv(out_file, sep='\t', header=False, index=False)


def read_filter_write(df_iter, f_filter, out_filename):
    """ Write filtered set of streamed data.

    Args:
        df_iter (iter of pandas.DataFrame): streamed data
        f_filter (callable): filtering function
        out_filename (str): output tsv filename

    """

    with open(out_filename, 'w') as out_file:
        for chunk in df_iter:
            chunk = f_filter(chunk)
            chunk.to_csv(out_file, sep='\t', header=False, index=False)


def group_aware_iter(df_iter, group_cols):
    """ Stream data on group boundaries.

    Args:
        df_iter (iter of pandas.DataFrame): streamed data
        group_cols (list of str): columns defining group

    Yields:
        pandas.DataFrame: group aware data

    """

    prev_data = None

    for df in df_iter:

        if len(df.index) == 0:
            return

        last_group = df.loc[df.index[-1], group_cols].values

        # Remove last group
        next_data = df.loc[(df[group_cols] != last_group).all(axis=1)]

        # Add previous last group to beginning of data
        if prev_data is not None:
            next_data = pd.concat([prev_data, next_data], ignore_index=True)

        yield next_data

        # Save last group for next yield
        prev_data = df.loc[(df[group_cols] == last_group).all(axis=1)]

    yield prev_data





import os
import shutil

import pypeliner
import pypeliner.managed as mgd

import destruct.benchmark.wrappers.tasks
import destruct.benchmark.wrappers.delly.tasks
import destruct.benchmark.wrappers.utils


def create_delly_wrapper_workflow(bam_filenames, output_filename, raw_data_dir, control_id=None, ref_genome_fasta_file=None, delly_excl_chrom=None):
    bams = list()
    for lib_id, bam_filename in bam_filenames.items():
        bams += [destruct.benchmark.wrappers.utils.symlink(bam_filename, link_name='{0}.bam'.format(lib_id), link_directory=raw_data_dir)]
        destruct.benchmark.wrappers.utils.symlink(bam_filename+'.bai', link_name='{0}.bam.bai'.format(lib_id), link_directory=raw_data_dir)

    workflow = pypeliner.workflow.Workflow()
    
    workflow.transform(
        name='get_sv_types',
        func=destruct.benchmark.wrappers.delly.tasks.get_sv_types,
        ret=pypeliner.managed.OutputChunks('sv_type'),
        args=(
            mgd.InputFile(ref_genome_fasta_file),
        ),
    )

    workflow.transform(
        name='delly_call',
        axes=('sv_type',),
        ctx={'mem': 64, 'num_retry': 2, 'mem_retry_factor': 2},
        func=destruct.benchmark.wrappers.delly.tasks.run_delly_call,
        args=(
            mgd.Instance('sv_type'),
            delly_excl_chrom,
            ref_genome_fasta_file,
            [mgd.InputFile(bam) for bam in bams],
            mgd.TempOutputFile('out.bcf', 'sv_type'),
        ),
    )

    if control_id is None:
        concat_input = mgd.TempInputFile('out.bcf', 'sv_type')

    else:
        workflow.transform(
            name='delly_filter_somatic',
            axes=('sv_type',),
            ctx={'mem': 4, 'num_retry': 2, 'mem_retry_factor': 2},
            func=destruct.benchmark.wrappers.delly.tasks.run_delly_filter,
            args=(
                mgd.Instance('sv_type'),
                bam_filenames.keys(),
                control_id, 
                mgd.TempSpace('samples.tsv'),
                ref_genome_fasta_file,
                mgd.TempInputFile('out.bcf', 'sv_type'),
                mgd.TempOutputFile('somatic.bcf', 'sv_type'),
            ),
        )

        concat_input = mgd.TempInputFile('somatic.bcf', 'sv_type')

    workflow.transform(
        name='concatenate_vcf',
        func=destruct.benchmark.wrappers.tasks.concatenate_bcf,
        ctx={'mem': 4, 'num_retry': 2, 'mem_retry_factor': 2},
        args=(
            concat_input,
            mgd.TempOutputFile('somatic.bcf'),
        ),
    )

    workflow.transform(
        name='convert_vcf',
        func=destruct.benchmark.wrappers.delly.tasks.convert_vcf,
        ctx={'mem': 4, 'num_retry': 3, 'mem_retry_increment': 2},
        args=(
            mgd.TempInputFile('somatic.bcf'),
            mgd.OutputFile(output_filename),
        ),
        kwargs={
            'control_id': control_id,
        }
    )

    return workflow


def setup_delly(test_config, **kwargs):
    ensembl_version = kwargs.get('ensembl_version', '70')
    ensembl_genome_version = kwargs.get('ensembl_genome_version', 'GRCh37')

    primary_assembly_url = 'ftp://ftp.ensembl.org/pub/release-'+ensembl_version+'/fasta/homo_sapiens/dna/Homo_sapiens.'+ensembl_genome_version+'.'+ensembl_version+'.dna.primary_assembly.fa.gz'
    chromosome_url = 'ftp://ftp.ensembl.org/pub/release-'+ensembl_version+'/fasta/homo_sapiens/dna/Homo_sapiens.'+ensembl_genome_version+'.'+ensembl_version+'.dna.chromosome.{0}.fa.gz'
    hg19_excl_url = 'http://raw.githubusercontent.com/tobiasrausch/delly/master/excludeTemplates/human.hg19.excl.tsv'

    if ensembl_genome_version != 'GRCh37':
        raise Exception('Only hg19 supported')
    
    genome_fasta = kwargs['ref_genome_fasta_file']
    delly_excl_chrom = kwargs['delly_excl_chrom']

    destruct.benchmark.wrappers.utils.makedirs(os.path.dirname(genome_fasta))
    destruct.benchmark.wrappers.utils.makedirs(os.path.dirname(delly_excl_chrom))

    if test_config.get('chromosomes', None) is None:
        destruct.benchmark.wrappers.utils.wget_file_gunzip(primary_assembly_url, genome_fasta)

    else:
        with open(genome_fasta, 'w') as genome_file:
            for chromosome in test_config['chromosomes']:
                chromosome_filename = './chromosome_{0}.fa'.format(chromosome)

                destruct.benchmark.wrappers.utils.wget_file_gunzip(chromosome_url.format(chromosome), chromosome_filename)

                with open(chromosome_filename, 'r') as chromosome_file:
                    shutil.copyfileobj(chromosome_file, genome_file)

                os.remove(chromosome_filename)

    destruct.benchmark.wrappers.utils.wget_file(hg19_excl_url, delly_excl_chrom)



import os
import argparse
import yaml

import pypeliner
import pypeliner.workflow
import pypeliner.managed as mgd

import destruct.benchmark.align.bwa.workflow
import destruct.benchmark.destruct_test
import destruct.benchmark.create_breakpoint_simulation


if __name__ == '__main__':

    argparser = argparse.ArgumentParser()
    pypeliner.app.add_arguments(argparser)

    argparser.add_argument('sim_config',
                           help='Simulation configuration filename')

    argparser.add_argument('ref_data_dir',
                           help='Reference genomes directory')

    argparser.add_argument('bam',
                           help='Source bam filename')

    argparser.add_argument('ref',
                           help='Reference genome for source bam')

    argparser.add_argument('results_dir',
                           help='Output directory')

    argparser.add_argument('tool_defs',
                           help='Tool Definition Filename')

    args = vars(argparser.parse_args())

    try:
        os.makedirs(args['results_dir'])
    except OSError:
        pass

    args['tmpdir'] = os.path.join(args['results_dir'], 'tmp')

    pyp = pypeliner.app.Pypeline(config=args)

    yaml_text = open(args['tool_defs']).read().format(ref_data_dir=args['ref_data_dir'])
    tool_defs = yaml.load(yaml_text)

    sim_config = yaml.load(open(args['sim_config']))

    ctx = {'mem':4}

    workflow = pypeliner.workflow.Workflow(default_ctx=ctx)

    workflow.setobj(mgd.TempOutputObj('simulation.params'), sim_config['simulation'])

    if sim_config['reference'].get('chromosomes', None) is not None:
        genome_fasta = os.path.join(args['results_dir'], 'genome.fa')
        source_bam = os.path.join(args['results_dir'], 'source.bam')

        workflow.setobj(mgd.TempOutputObj('chromosomes'), sim_config['reference']['chromosomes'])

        workflow.transform(
            name='create_ref_bam',
            func=destruct.benchmark.destruct_test.create_ref_bam,
            args=(
                mgd.InputFile(args['ref']),
                mgd.InputFile(args['bam']),
                mgd.OutputFile(genome_fasta),
                mgd.OutputFile(source_bam),
                mgd.TempInputObj('chromosomes'),
            ),
        )
    
    else:
        genome_fasta = args['ref']
        source_bam = args['bam']

    workflow.transform(
        name='create_sim',
        func=destruct.benchmark.create_breakpoint_simulation.create_breakpoints,
        args=(
            mgd.TempInputObj('simulation.params'),
            mgd.InputFile(genome_fasta),
            mgd.OutputFile(os.path.join(args['results_dir'], 'simulated.fa')),
            mgd.OutputFile(os.path.join(args['results_dir'], 'simulated.tsv')),
        ),
    )

    workflow.transform(
        name='partition',
        func=destruct.benchmark.destruct_test.partition_bam,
        args=(
            mgd.InputFile(source_bam),
            mgd.TempOutputFile('normal.raw.bam'),
            mgd.TempOutputFile('tumour.unspiked.bam'),
            0.5,
        ),
    )

    workflow.commandline(
        name='simulate',
        args=(
            'destruct_bamextractsimreads',
            '-b', mgd.InputFile(source_bam),
            '-r', mgd.InputFile(genome_fasta),
            '-s', mgd.InputFile(os.path.join(args['results_dir'], 'simulated.fa')),
            '-f', mgd.TempInputObj('simulation.params').extract(lambda a: a['coverage_fraction']),
            '-1', mgd.OutputFile(os.path.join(args['results_dir'], 'simulated.1.fastq')),
            '-2', mgd.OutputFile(os.path.join(args['results_dir'], 'simulated.2.fastq')),
        ),
    )

    workflow.subworkflow(
        name='bwa_align',
        func=destruct.benchmark.align.bwa.workflow.bwa_align_workflow,
        args=(
            mgd.InputFile(genome_fasta),
            mgd.InputFile(os.path.join(args['results_dir'], 'simulated.1.fastq')),
            mgd.InputFile(os.path.join(args['results_dir'], 'simulated.2.fastq')),
            mgd.TempOutputFile('simulated.unsorted.bam'),
        ),
        kwargs={
            'read_group_str': '@RG\\tID:B',
        },
    )

    workflow.transform(
        name='samtools_merge_sort_index',
        func=destruct.benchmark.destruct_test.samtools_merge_sort_index,
        args=(
            mgd.TempOutputFile('tumour.raw.bam'),
            mgd.TempInputFile('tumour.unspiked.bam'),
            mgd.TempInputFile('simulated.unsorted.bam'),
        ),
    )

    workflow.transform(
        name='samtools_reheader_tumour',
        func=destruct.benchmark.destruct_test.samtools_sample_reheader,
        args=(
            mgd.OutputFile(os.path.join(args['results_dir'], 'tumour.bam')),
            mgd.TempInputFile('tumour.raw.bam'),
            'tumour',
        ),
    )

    workflow.transform(
        name='samtools_reheader_normal',
        func=destruct.benchmark.destruct_test.samtools_sample_reheader,
        args=(
            mgd.OutputFile(os.path.join(args['results_dir'], 'normal.bam')),
            mgd.TempInputFile('normal.raw.bam'),
            'normal',
        ),
    )

    workflow.setobj(
        obj=mgd.TempOutputObj('tool_defs', 'tool_name'),
        value=tool_defs,
    )

    workflow.subworkflow(
        name='run_tool',
        axes=('tool_name',),
        func=destruct.benchmark.destruct_test.create_tool_workflow,
        args=(
            mgd.TempInputObj('tool_defs', 'tool_name'),
            {
                'tumour': mgd.InputFile(os.path.join(args['results_dir'], 'tumour.bam')),
                'normal': mgd.InputFile(os.path.join(args['results_dir'], 'normal.bam')),
            },
            mgd.OutputFile(os.path.join(args['results_dir'], 'results_{tool_name}.tsv'), 'tool_name'),
            mgd.TempSpace('tool_raw_data', 'tool_name', cleanup='none'),
        ),
        kwargs={
            'control_id': 'normal',
        },
    )

    workflow.transform(
        name='plot',
        axes=('tool_name',),
        func=destruct.benchmark.destruct_test.create_roc_plot,
        args=(
            mgd.TempInputObj('simulation.params'),
            mgd.TempInputObj('tool_defs', 'tool_name'),
            mgd.InputFile(genome_fasta),
            mgd.InputFile(os.path.join(args['results_dir'], 'simulated.tsv')),
            mgd.InputFile(os.path.join(args['results_dir'], 'results_{tool_name}.tsv'), 'tool_name'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'annotated_{tool_name}.tsv'), 'tool_name'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'identified_{tool_name}.tsv'), 'tool_name'),
            mgd.OutputFile(os.path.join(args['results_dir'], 'plots_{tool_name}.pdf'), 'tool_name'),
        ),
    )

    pyp.run(workflow)





import argparse

import destruct.ui.run
import destruct.ui.create_ref_data
import destruct.ui.extract_somatic


def main():
    argparser = argparse.ArgumentParser()
    subparsers = argparser.add_subparsers()
    
    destruct.ui.run.add_arguments(subparsers.add_parser('run'))
    destruct.ui.create_ref_data.add_arguments(subparsers.add_parser('create_ref_data'))
    destruct.ui.extract_somatic.add_arguments(subparsers.add_parser('extract_somatic'))

    args = vars(argparser.parse_args())
    func = args.pop('func')
    func(**args)



import pandas as pd

import destruct.balanced


def classify_rearrangement_type(entry):
    break1 = entry['position_1']
    break2 = entry['position_2']
    size = abs(int(break1) - int(break2))
    orientation_type = entry['type']
    if entry['balanced']:
        rearrangement_type = 'balanced'
    elif size <= 1000000 and orientation_type == 'deletion':
        rearrangement_type = 'deletion'
    elif size <= 10000 and orientation_type == 'inversion':
        rearrangement_type = 'foldback'
    elif size <= 1000000 and orientation_type == 'inversion':
        rearrangement_type = 'inversion'
    elif size <= 1000000 and orientation_type == 'duplication':
        rearrangement_type = 'duplication'
    else:
        rearrangement_type = 'unbalanced'
    return rearrangement_type


def annotate_rearrangement_type(df):
    rearrangement_types = list()
    for idx, row in df.iterrows():
        rearrangement_types.append(classify_rearrangement_type(row))
    df['rearrangement_type'] = rearrangement_types


def filter_annotate_breakpoints(
        input_breakpoint_filename,
        input_breakpoint_library_filename,
        control_ids,
        output_breakpoint_filename,
        output_breakpoint_library_filename,
        patient_library_filename=None):
    """ Filter and annotate breakpoints.

    Args:
        input_breakpoint_filename (str): filename of breakpoint table
        input_breakpoint_library_filename (str): filename of breakpoint library table
        control_ids (list): control id or ids
        output_breakpoint_filename (str): output filename of breakpoint table
        output_breakpoint_library_filename (str): output filename of breakpoint library table

    KwArgs:
        patient_libraries (str): dataframe of library ids for each patient, columns patient_id, library

    If patient_libraries is not specified, assumed one patient for all libraries.

    """

    brk = pd.read_csv(input_breakpoint_filename, sep='\t',
                      converters={'chromosome_1': str, 'chromosome_2': str})

    brklib = pd.read_csv(input_breakpoint_library_filename, sep='\t')

    if patient_library_filename is None:
        patient_libraries = {'null': list(brklib['library'].unique())}

    else:
        patient_library_data = pd.read_csv(patient_library_filename, sep='\t')
        patient_libraries = {}
        for patient_id, rows in patient_library_data.groupby('patient_id'):
            patient_libraries[patient_id] = [row['library'] for row in rows]

    # Add is_normal column
    brklib['is_normal'] = brklib['library'].isin(control_ids)

    # Add patient_id column
    brklib['patient_id'] = ''
    for patient_id, library_ids in patient_libraries.items():
        for library_id in library_ids:
            brklib.loc[brklib['library'] == library_id, 'patient_id'] = patient_id

    num_patients = (
        brklib[['prediction_id', 'patient_id']]
        .drop_duplicates()
        .groupby('prediction_id')
        .size()
        .rename('num_patients')
    )

    # Mark as germline any prediction with nonzero normal reads
    is_germline = brklib[brklib['num_reads'] > 0].groupby('prediction_id')['is_normal'].any()

    # DGV predictions also germline
    is_dgv = brk.set_index('prediction_id')['dgv_ids'].notnull()

    brk.set_index('prediction_id', inplace=True)
    brk['is_germline'] = is_germline
    brk['is_dgv'] = is_dgv
    brk['num_patients'] = num_patients
    brk.reset_index(inplace=True)

    # Mark as a filtered any breakpoint that is common or germline
    brk['is_filtered'] = brk['is_germline'] | brk['is_dgv'] | (brk['num_patients'] > 1)

    # Get a list of breakends and their filtered status
    def get_brkend(brk, side, data_cols):
        cols = ['chromosome', 'strand', 'position']
        side_cols = [a + '_' + side for a in cols]
        brkend = brk[['prediction_id'] + side_cols + data_cols]
        brkend = brkend.rename(columns=dict(zip(side_cols, cols)))
        brkend['side'] = side
        return brkend

    brkend = pd.concat([get_brkend(brk, '1', ['is_filtered']),
                        get_brkend(brk, '2', ['is_filtered'])], ignore_index=True)

    # Calculate the minimum distance to the nearest filtered breakend for all breakpoints
    def calculate_dist_filtered(data):

        data = data.sort_values('position')

        data['nearest_left'] = None
        data['nearest_right'] = None

        data.loc[data['is_filtered'], 'nearest_left'] = data.loc[data['is_filtered'], 'position']
        data.loc[data['is_filtered'], 'nearest_right'] = data.loc[data['is_filtered'], 'position']

        data['nearest_left'] = data['nearest_left'].fillna(method='ffill')
        data['nearest_right'] = data['nearest_right'].fillna(method='bfill')

        data['nearest_left'] = data['position'] - data['nearest_left']
        data['nearest_right'] = data['nearest_right'] - data['position']

        data['dist_filtered'] = data[['nearest_left', 'nearest_right']].min(axis=1)

        return data['dist_filtered']

    dist_filtered = brkend.set_index(['prediction_id', 'side'])\
                          .groupby(['chromosome', 'strand'])\
                          .apply(calculate_dist_filtered)\
                          .reset_index(level=[0, 1], drop=True)\
                          .unstack(level=1)\
                          .min(axis=1)

    brk.set_index('prediction_id', inplace=True)
    brk['dist_filtered'] = dist_filtered
    brk.reset_index(inplace=True)

    chromosomes = ['1', '2', '3', '4', '5', '6', '7', '8', '9', '10',
                   '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22',
                   'X', 'Y']

    # Filter breakpoints
    brk = brk.loc[(~brk['is_filtered']) &
                  (brk['dist_filtered'] > 50) &
                  (brk['num_reads'] >= 2) &
                  (brk['num_split'] >= 2) &
                  (brk['log_likelihood'] > -20.) &
                  (brk['template_length_min'] > 120) &
                  (brk['chromosome_1'].isin(chromosomes)) &
                  (brk['chromosome_2'].isin(chromosomes))]

    brklib = brklib.merge(brk[['prediction_id']].drop_duplicates())

    # Balanced rearrangement annotation
    balanced_prediction_ids = []

    for patient_id, library_ids in patient_libraries.items():
        patient_prediction_ids = brklib.loc[brklib['library'].isin(library_ids), ['prediction_id']].drop_duplicates()
        patient_brks = brk.merge(patient_prediction_ids)
        balanced_rearrangements = destruct.balanced.detect_balanced_rearrangements(patient_brks)
        for rearrangement in balanced_rearrangements:
            balanced_prediction_ids.extend(rearrangement.prediction_ids)

    brk['balanced'] = False
    brk.loc[brk['prediction_id'].isin(balanced_prediction_ids), 'balanced'] = True

    # Annotate rearrangement type
    annotate_rearrangement_type(brk)

    # Output filtered breakpoint tables
    brk.to_csv(output_breakpoint_filename, sep='\t', header=True, index=False)
    brklib.to_csv(output_breakpoint_library_filename, sep='\t', header=True, index=False)





import os
import pypeliner

def wget_gunzip(url, filename):
    temp_filename = filename + '.tmp'
    pypeliner.commandline.execute('wget', url, '-c', '-O', temp_filename + '.gz')
    pypeliner.commandline.execute('gunzip', temp_filename + '.gz')
    os.rename(temp_filename, filename)


def wget(url, filename):
    temp_filename = filename + '.tmp'
    pypeliner.commandline.execute('wget', url, '-c', '-O', temp_filename)
    os.rename(temp_filename, filename)


# REFACTOR: repeated in wrappers.utils
def remove(filename):
    try:
        os.remove(filename)
    except OSError as e:
        if e.errno != 2:
            raise


default_chromosomes = ('1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', 'MT')

ensembl_version = '70'
ensembl_genome_version = 'GRCh37'
default_ensembl_assembly_url = 'ftp://ftp.ensembl.org/pub/release-'+ensembl_version+'/fasta/homo_sapiens/dna/Homo_sapiens.'+ensembl_genome_version+'.'+ensembl_version+'.dna.{0}.fa.gz'


def download_genome_fasta(genome_fasta,
                          chromosomes=default_chromosomes,
                          include_nonchromosomal=True,
                          assembly_url=default_ensembl_assembly_url):

    ensembl_assemblies = ['chromosome.'+a for a in chromosomes]

    if include_nonchromosomal:
        ensembl_assemblies += ['nonchromosomal']

    temp_filenames = list()

    with open(genome_fasta, 'w') as genome_file:

        for assembly in ensembl_assemblies:

            assembly_fasta = genome_fasta + '.dna.assembly.{0}.fa'.format(assembly)

            if not os.path.exists(assembly_fasta):
                wget_gunzip(assembly_url.format(assembly), assembly_fasta)

            with open(assembly_fasta, 'r') as assembly_file:
                for line in assembly_file:
                    if line[0] == '>':
                        line = line.split()[0] + '\n'
                    genome_file.write(line)

            temp_filenames.append(assembly_fasta)

    for temp_filename in temp_filenames:
        remove(temp_filename)





import argparse
import destruct.create_ref_data


def create_ref_data(**args):
    config = {}
    if args['config'] is not None:
        exec(open(args['config']).read(), {}, config)

    destruct.create_ref_data.create_ref_data(config, args['ref_data_dir'])


def add_arguments(argparser):
    argparser.add_argument('ref_data_dir',
                           help='Reference dataset directory')

    argparser.add_argument('-c', '--config',
                           help='Configuration filename')

    argparser.set_defaults(func=create_ref_data)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser()

    add_arguments(argparser)

    args = vars(argparser.parse_args())
    func = args.pop('func')
    func(**args)



import collections
import itertools
import numpy as np
import pandas as pd

import networkx
import blossomv.blossomv


BalancedRearrangment = collections.namedtuple(
    'BalancedRearrangment',
    [
        'deleted_length',
        'duplicated_length',
        'prediction_ids',
    ],
)


def detect_balanced_rearrangements(
    breakpoints,
    dec_nt_per_break=2000.,
    inc_nt_per_break=500.,
    cost_resolution=1000.,
):
    if len(breakpoints.index) == 0:
        return []

    G = networkx.Graph()

    breakpoints = breakpoints.copy()

    # Algorithm cannot handle loops created by perfect foldbacks
    # for these events, move one breakpoint by one nucleotide
    is_perfect_foldback = (
        (breakpoints['chromosome_1'] == breakpoints['chromosome_2']) &
        (breakpoints['position_1'] == breakpoints['position_2']) &
        (breakpoints['strand_1'] == breakpoints['strand_2']))
    breakpoints.loc[is_perfect_foldback, 'position_2'] += 1

    break_ends = pd.DataFrame(
        {
            'chromosome': np.concatenate([breakpoints['chromosome_1'].values, breakpoints['chromosome_2'].values]),
            'position': np.concatenate([breakpoints['position_1'].values, breakpoints['position_2'].values]),
            'strand': np.concatenate([breakpoints['strand_1'].values, breakpoints['strand_2'].values]),
        }
    )
    break_ends.drop_duplicates(inplace=True)

    # Add break end nodes
    for idx in break_ends.index:
        break_end = (
            break_ends.loc[idx, 'chromosome'],
            break_ends.loc[idx, 'position'],
            break_ends.loc[idx, 'strand'],
        )
        G.add_node(break_end)

    # Add breakpoint edges
    for idx in breakpoints.index:
        break_end_1 = (
            breakpoints.loc[idx, 'chromosome_1'],
            breakpoints.loc[idx, 'position_1'],
            breakpoints.loc[idx, 'strand_1'],
        )
        break_end_2 = (
            breakpoints.loc[idx, 'chromosome_2'],
            breakpoints.loc[idx, 'position_2'],
            breakpoints.loc[idx, 'strand_2'],
        )
        prediction_id = breakpoints.loc[idx, 'prediction_id']
        G.add_edge(break_end_1, break_end_2, edge_type='breakpoint', prediction_id=prediction_id)

    # Add reference and segment edges
    for chromosome, chrom_break_ends in break_ends.groupby('chromosome'):
        chrom_break_ends = chrom_break_ends.sort_values('position')

        # Add reference edges
        for position in chrom_break_ends['position'].values:
            break_end_1 = (chromosome, position, '+')
            break_end_2 = (chromosome, position, '-')
            G.add_edge(break_end_1, break_end_2, edge_type='reference')

        # Add segment edges
        for start, end in zip(chrom_break_ends['position'].values[:-1], chrom_break_ends['position'].values[1:]):
            break_end_1 = (chromosome, start, '-')
            break_end_2 = (chromosome, end, '+')
            length = end - start
            G.add_edge(break_end_1, break_end_2, edge_type='segment', length=length)

    # Create a genome modification graph
    #  - identical node set as genome graph
    #  - create a +1 and -1 signed edge for each original edge
    #  - color each edge:
    #     - segment edges: +1 -> red, -1 -> blue
    #     - bond edges: +1 -> blue, -1 -> red
    #  - add a cost for each edge:
    #     - segment edges: cost for increasing / decreasing per nt
    #     - breakpoint edges: decreasing inf, increasing -1

    # Create graph with duplicated edges for each sign
    H = networkx.MultiGraph()
    H.add_nodes_from(G)
    for edge in G.edges():
        H.add_edge(*edge, attr_dict=G.get_edge_data(*edge), sign=1)
        H.add_edge(*edge, attr_dict=G.get_edge_data(*edge), sign=-1)

    # Color each edge, +1 for red, -1 for blue
    # add cost for each edge
    for edge in H.edges():
        for multi_edge_idx, edge_attr in H[edge[0]][edge[1]].items():
            edge_type = edge_attr['attr_dict']['edge_type']
            edge_attr['cost'] = 0
            if edge_type == 'segment':
                edge_attr['color'] = edge_attr['sign']
                if edge_attr['sign'] == 1:
                    edge_attr['cost'] = float(edge_attr['attr_dict']['length'] / inc_nt_per_break)
                elif edge_attr['sign'] == -1:
                    edge_attr['cost'] = float(edge_attr['attr_dict']['length'] / dec_nt_per_break)
            elif edge_type == 'breakpoint':
                edge_attr['color'] = -edge_attr['sign']
                if edge_attr['sign'] == 1:
                    edge_attr['cost'] = -1.
                elif edge_attr['sign'] == -1:
                    edge_attr['cost'] = np.inf
            elif edge_type == 'reference':
                edge_attr['color'] = -edge_attr['sign']
                edge_attr['cost'] = 1. / cost_resolution
            else:
                raise ValueError('unknown edge type {}'.format(edge_type))

    # Create matching graph
    #  - duplicate nodes, one set red, one set blue
    #  - add transverse edges (v_red, v_blue)
    #  - for each original edge:
    #    - add (u_red, v_red) for red edges
    #    - add (u_blue, v_blue) for blue edges
    #    - replacate edge costs

    transverse_edge_cost = 1. / cost_resolution

    M = networkx.Graph()
    for node in H.nodes():
        transverse_edge = []
        for color in (1, -1):
            colored_node = node + (color,)
            M.add_node(colored_node)
            transverse_edge.append(colored_node)
        M.add_edge(*transverse_edge, cost=transverse_edge_cost)

    for edge in H.edges():
        for multi_edge_idx, edge_attr in H[edge[0]][edge[1]].items():
            cost = edge_attr['cost']
            if np.isinf(cost):
                continue
            color = edge_attr['color']
            colored_node_1 = edge[0] + (color,)
            colored_node_2 = edge[1] + (color,)
            M.add_edge(colored_node_1, colored_node_2, attr_dict=edge_attr, cost=cost)

    M1 = networkx.convert_node_labels_to_integers(M, label_attribute='node_tuple')

    # Min cost perfect matching
    edges = networkx.get_edge_attributes(M1, 'cost')
    for edge in edges.keys():
        if edge[0] == edge[1]:
            raise Exception('self loop {}'.format(M1[edge[0]][edge[1]]))
        edges[edge] = int(edges[edge] * cost_resolution)
    min_cost_edges = blossomv.blossomv.min_weight_perfect_matching(edges)

    # Remove unselected edges
    assert set(min_cost_edges).issubset(edges.keys())
    remove_edges = set(edges.keys()).difference(min_cost_edges)
    M2 = M1.copy()
    M2.remove_edges_from(remove_edges)

    # Re-create original graph with matched edges
    M3 = networkx.relabel_nodes(M2, mapping=networkx.get_node_attributes(M2, 'node_tuple'))

    # Create subgraph of H with only selected edges
    H1 = networkx.Graph()
    for edge in M3.edges():
        edge_attr = M3[edge[0]][edge[1]]
        node_1 = edge[0][:-1]
        node_2 = edge[1][:-1]
        if node_1 == node_2:
            continue
        if H1.has_edge(node_1, node_2):
            H1.remove_edge(node_1, node_2)
        else:
            H1.add_edge(node_1, node_2, attr_dict=edge_attr)

    # Get individual events
    rearrangements = []
    for C in networkx.connected_component_subgraphs(H1, copy=False):
        duplicated_length = 0
        deleted_length = 0
        prediction_ids = []

        for edge in C.edges():
            edge_type = C[edge[0]][edge[1]]['attr_dict']['attr_dict']['attr_dict']['edge_type']

            if edge_type == 'segment':
                sign = C[edge[0]][edge[1]]['attr_dict']['attr_dict']['sign']
                length = C[edge[0]][edge[1]]['attr_dict']['attr_dict']['attr_dict']['length']
                if sign == 1:
                    duplicated_length += length
                else:
                    deleted_length += length

            elif edge_type == 'breakpoint':
                prediction_id = C[edge[0]][edge[1]]['attr_dict']['attr_dict']['attr_dict']['prediction_id']
                prediction_ids.append(prediction_id)

        if len(prediction_ids) <= 1:
            continue

        rearrangements.append(
            BalancedRearrangment(
                deleted_length,
                duplicated_length,
                prediction_ids,
            )
        )

    return rearrangements




import gzip
import itertools
import os

import pypeliner


def split_fastq(in_filename, num_reads_per_file, out_filename_callback):
    if in_filename.endswith('.gz'):
        opener = gzip.open
    else:
        opener = open
    with opener(in_filename, 'r') as in_file:
        file_number = 0
        out_file = None
        out_file_read_count = None
        try:
            for name, seq, comment, qual in itertools.zip_longest(*[in_file]*4):
                if out_file is None or out_file_read_count == num_reads_per_file:
                    if out_file is not None:
                        out_file.close()
                    out_file = open(out_filename_callback(file_number), 'w')
                    out_file_read_count = 0
                    file_number += 1
                out_file.write(name)
                out_file.write(seq)
                out_file.write(comment)
                out_file.write(qual)
                out_file_read_count += 1
        finally:
            if out_file is not None:
                out_file.close()


def sort_bam(input_bam, output_bam):
    pypeliner.commandline.execute('samtools', 'sort', '-o', input_bam, output_bam+'.sortprefix', '>', output_bam)


def split_dict(d):
    ditems = d.items()
    return dict(ditems[:len(ditems)/2]), dict(ditems[len(ditems)/2:])


def cat_merge(in_filenames, out_filename):
    in_filename_args = list((a[1] for a in sorted(in_filenames.items())))
    cat_args = ['cat'] + in_filename_args + ['>', out_filename]
    pypeliner.commandline.execute(*cat_args)


def merge_bam(input_bams, output_bam):
    if len(input_bams) >= 100:
        output_bam_tmp1 = output_bam + ".tmp1"
        output_bam_tmp2 = output_bam + ".tmp2"
        try: os.remove(output_bam_tmp1)
        except: pass
        try: os.remove(output_bam_tmp2)
        except: pass
        input_bams_1, input_bams_2 = split_dict(input_bams)
        merge_bam(input_bams_1, output_bam_tmp1)
        merge_bam(input_bams_2, output_bam_tmp2)
        merge_bam({'tmp1':output_bam_tmp1, 'tmp2':output_bam_tmp2}, output_bam)
        os.remove(output_bam_tmp1)
        os.remove(output_bam_tmp2)
    elif len(input_bams) == 1:
        pypeliner.commandline.execute('cp', list(input_bams.values())[0], output_bam)
    else:
        pypeliner.commandline.execute('samtools', 'merge', '-c', output_bam, *input_bams.values())



import os

import pypeliner.workflow
import pypeliner.managed as mgd

import destruct.benchmark.align.bwa.tasks


def bwa_align_workflow(
    genome_fasta_filename,
    fastq_1_filename,
    fastq_2_filename,
    bam_filename,
    reads_per_job=int(1e6),
    read_group_str=None,
):
    if read_group_str is None:
        read_group_str = '@RG\\tID:S1\\tSM:sample_1'

    lowmem = {'mem':1}
    himem = {'mem':8}

    if not os.path.exists(genome_fasta_filename+'.bwt'):
        raise Exception('No index for ' + genome_fasta_filename)

    workflow = pypeliner.workflow.Workflow()

    workflow.transform(
        name='split1',
        ctx=lowmem,
        func=destruct.benchmark.align.bwa.tasks.split_fastq,
        args=(
            mgd.InputFile(fastq_1_filename),
            reads_per_job,
            mgd.TempOutputFile('fastq1', 'byread'),
        ),
    )

    workflow.transform(
        name='split2',
        ctx=lowmem,
        func=destruct.benchmark.align.bwa.tasks.split_fastq,
        args=(
            mgd.InputFile(fastq_2_filename), 
            reads_per_job, 
            mgd.TempOutputFile('fastq2', 'byread', axes_origin=[]),
        ),
    )

    workflow.commandline(
        name='align',
        axes=('byread',),
        ctx=himem,
        args=(
            'bwa', 'mem',
            '-R', read_group_str,
            genome_fasta_filename,
            mgd.TempInputFile('fastq1', 'byread'),
            mgd.TempInputFile('fastq2', 'byread'),
            '|',
            'samtools', 'view', '-bt',
            genome_fasta_filename+'.fai',
            '-',
            '-o', mgd.TempOutputFile('bam', 'byread'),
        ),
    )

    workflow.transform(
        name='merge_bams',
        ctx=lowmem,
        func=destruct.benchmark.align.bwa.tasks.merge_bam,
        args=(
            mgd.TempInputFile('bam', 'byread'),
            mgd.OutputFile(bam_filename),
        ),
    )

    return workflow





import os
import pysam
import pandas as pd

import pypeliner.commandline


def _rename_index(out_file):
    if out_file.endswith('.tmp'):
        out_index = out_file + '.csi'
        renamed_out_index = out_file[:-4] + '.csi'
        try:
            os.remove(renamed_out_index)
        except OSError:
            pass
        os.rename(out_index, renamed_out_index)


def _write_empty_vcf(out_file, bam_files):
    with open(out_file+'.vcf', 'w') as f:
        f.write('##fileformat=VCFv4.2\n')
        f.write('#CHROM\tPOS\tID\tREF\tALT\tQUAL\tFILTER\tINFO\tFORMAT')
        for bam_file in bam_files:
            sample_id = os.path.splitext(os.path.basename(bam_file))[0]
            f.write('\t{}'.format(sample_id))
        f.write('\n')

    pypeliner.commandline.execute('bcftools', 'convert', '-O', 'b', out_file+'.vcf', '-o', out_file)
    pypeliner.commandline.execute('bcftools', 'index', out_file)
    os.remove(out_file+'.vcf')


def get_sv_types(ref_genome_fasta_file):

    # Check for a genome with a single chromosome
    num_chromosomes = 0
    with open(ref_genome_fasta_file) as f:
        for line in f:
            if line.startswith('>'):
                num_chromosomes += 1
            if num_chromosomes > 1:
                break

    # Translocations only for genomes with multiple chromosomes
    if num_chromosomes > 1:
        return ('DEL', 'DUP', 'INV', 'TRA', 'INS')
    else:
        return ('DEL', 'DUP', 'INV', 'INS')


def run_delly_call(sv_type, delly_excl_chrom, ref_genome_fasta_file, bam_files, out_file):
    delly_args = [
        'delly', 'call',
        '-t', sv_type,
        '-x', delly_excl_chrom,
        '-g', ref_genome_fasta_file,
        '-o', out_file,
    ]

    delly_args += bam_files
    
    pypeliner.commandline.execute(*delly_args)

    if not os.path.exists(out_file):
        _write_empty_vcf(out_file, bam_files)

    _rename_index(out_file)


def run_delly_filter(sv_type, sample_ids, control_id, sample_file, ref_genome_fasta_file, in_file, out_file):
    with open(sample_file, 'w') as f:
        for sample_id in sample_ids:
            if control_id == sample_id:
                sample_type = 'control'
            else:
                sample_type = 'tumor'
            f.write('{}\t{}\n'.format(sample_id, sample_type))

    delly_args = [
        'delly', 'filter',
        '-t', sv_type,
        '-f', 'somatic',
        '-o', out_file,
        '-s', sample_file,
        '-g', ref_genome_fasta_file,
        in_file,
    ]

    pypeliner.commandline.execute(*delly_args)
    _rename_index(out_file)


def convert_vcf(bcf_filename, output_filename, control_id=None):
    bcf_reader = pysam.VariantFile(bcf_filename, 'rb')

    breakpoint_table = list()
    breakpoint_library_table = list()

    for row in bcf_reader:
        prediction_id = row.id

        chrom_1 = row.chrom
        chrom_2 = row.info['CHR2']

        strand_1, strand_2 = [('-', '+')[a == '3'] for a in row.info['CT'].split('to')]

        coord_1 = row.pos
        coord_2 = row.info['END']

        if 'LowQual' in row.filter:
            qual = 0
        else:
            qual = 1

        breakpoint_table.append((prediction_id, chrom_1, chrom_2, strand_1, strand_2, coord_1, coord_2, qual))

        for sample, call in row.samples.items():
            num_spanning = call['DV']
            num_split = call['RV']

            breakpoint_library_table.append((prediction_id, sample, num_spanning, num_split))

    breakpoint_table = pd.DataFrame(
        breakpoint_table,
        columns=[
            'prediction_id',
            'chromosome_1', 'chromosome_2',
            'strand_1', 'strand_2',
            'position_1', 'position_2',
            'qual',
        ]
    )

    breakpoint_library_table = pd.DataFrame(
        breakpoint_library_table,
        columns=[
            'prediction_id',
            'library',
            'num_spanning',
            'num_split'
        ]
    )

    if len(breakpoint_table.index) == 0:
        raise Exception('No variants')

    spanning_read_counts = breakpoint_library_table.groupby('prediction_id')['num_spanning'].sum().reset_index()
    split_read_counts = breakpoint_library_table.groupby('prediction_id')['num_split'].sum().reset_index()

    total_read_counts = breakpoint_library_table.set_index(['prediction_id', 'library']).sum(axis=1).unstack()
    total_read_counts.columns = [a + '_count' for a in total_read_counts.columns]
    total_read_counts = total_read_counts.reset_index()

    breakpoint_table = breakpoint_table.merge(spanning_read_counts, on='prediction_id')
    breakpoint_table = breakpoint_table.merge(split_read_counts, on='prediction_id')
    breakpoint_table = breakpoint_table.merge(total_read_counts, on='prediction_id')

    # Filter based on evidence in control dataset
    if control_id is not None:
         breakpoint_table = breakpoint_table[breakpoint_table['{0}_count'.format(control_id)] == 0]

    breakpoint_table.to_csv(output_filename, sep='\t', index=False)







def read_sequences(fasta):
    id = None
    sequences = []
    for line in fasta:
        line = line.rstrip()
        if len(line) == 0:
            continue
        if line[0] == '>':
            if id is not None:
                yield (id, ''.join(sequences))
            id = line[1:].split()[0]
            sequences = []
        else:
            sequences.append(line)
    if id is not None:
        yield (id, ''.join(sequences))




import csv
import sys
import logging
import os
import ConfigParser
import re
import itertools
import subprocess
import argparse
import string
import collections
import pandas as pd

import pypeliner

if __name__ == '__main__':

    import realign_test
    from realign_test import *

    import destruct
    from destruct import *

    import create_breakpoint_simulation

    import score_stats

    parser = argparse.ArgumentParser()
    pypeliner.easypypeliner.add_arguments(parser)
    parser.add_argument('simconfig', help='Simulation Configuration filename')
    parser.add_argument('results', help='Results filename')
    args = parser.parse_args()

    cfg = pypeliner.easypypeliner.Config(vars(parser.parse_args()))
    pyp = pypeliner.easypypeliner.EasyPypeliner([destruct, realign_test], cfg)
    sch = pyp.sch

    lowmem = {'mem':1}
    medmem = {'mem':8}
    himem = {'mem':32}

    axes = ('bysim',)

    sch.transform('create', (), lowmem, create_simulation_params, sch.oobj('simulation.params', axes), sch.input(args.simconfig))

    sch.transform('generate', axes, medmem, create_breakpoint_simulation.create, None, cfg, sch.iobj('simulation.params', axes), sch.ofile('simulated.fasta', axes), sch.ofile('simulated.info', axes), sch.ofile('concordant1', axes), sch.ofile('concordant2', axes), sch.ofile('discordant1', axes), sch.ofile('discordant2', axes))

    sch.transform('index1', axes, lowmem, index_fastq, None, sch.ifile('concordant1', axes), sch.ofile('sample1', axes))
    sch.transform('index2', axes, lowmem, index_fastq, None, sch.ifile('concordant2', axes), sch.ofile('sample2', axes))
    sch.transform('index3', axes, lowmem, index_fastq, None, sch.ifile('discordant1', axes), sch.ofile('reads1', axes))
    sch.transform('index4', axes, lowmem, index_fastq, None, sch.ifile('discordant2', axes), sch.ofile('reads2', axes))

    sch.commandline('bwtsample', axes, medmem, cfg.bowtie2_bin, '--no-mixed', '--no-discordant', '--very-sensitive', '-X', cfg.fragment_length_max, cfg.bowtie2_options, '-x', cfg.genome_fasta, '-1', sch.ifile('sample1', axes), '-2', sch.ifile('sample2', axes), '>', sch.ofile('sample.sam', axes))
    sch.commandline('aligntrue', axes, medmem, cfg.aligntrue_tool, '-g', cfg.gap_score, '-x', cfg.mismatch_score, '-m', cfg.match_score, '-r', cfg.genome_fasta, '-a', sch.ifile('sample.sam', axes), '>', sch.ofile('samples.align.true', axes))
    sch.transform('scorestats', axes, medmem, score_stats.create_score_stats, None, sch.ifile('samples.align.true', axes), int(cfg.match_score), sch.ofile('score.stats', axes), sch.ofile('score.stats.plots', axes), sch.inst('bysim'))

    sch.transform('readstats', axes, medmem, calculate_concordant_stats, sch.oobj('stats', axes), sch.ifile('sample.sam', axes), float(cfg.fragment_length_num_stddevs))

    sch.transform('prepseed', axes, medmem, prepare_seed_fastq, None, sch.ifile('reads1', axes), sch.ifile('reads2', axes), 36, sch.ofile('reads.seed', axes))
    sch.commandline('prepreads', axes, medmem, 'cat', sch.ifile('reads1', axes), sch.ifile('reads2', axes), '>', sch.ofile('reads', axes))
    sch.commandline('bwtseed', axes, medmem, cfg.bowtie_bin, cfg.genome_fasta, sch.ifile('reads.seed', axes), '--chunkmbs', '512', '-k', '1000', '-m', '1000', '--strata', '--best', '-S', '>', sch.ofile('reads.seed.sam', axes))
    sch.commandline('realign', axes, medmem, cfg.realign2_tool, '-a', sch.ifile('reads.seed.sam', axes), '-s', sch.ifile('reads', axes), '-r', cfg.genome_fasta, '-g', cfg.gap_score, '-x', cfg.mismatch_score, '-m', cfg.match_score, '--flmin', sch.iobj('stats', axes).prop('fragment_length_min'), '--flmax', sch.iobj('stats', axes).prop('fragment_length_max'), '--tchimer', cfg.chimeric_threshold, '--talign', cfg.alignment_threshold, '--pchimer', cfg.chimeric_prior, '--tvalid', cfg.readvalid_threshold, '-z', sch.ifile('score.stats', axes), '--span', sch.ofile('spanning.alignments', axes), '--split', sch.ofile('split.alignments', axes))

    sch.transform('eval_span_align', axes, medmem, eval_span_align, None, sch.ifile('simulated.info', axes), sch.ifile('reads1', axes), sch.ifile('spanning.alignments', axes), sch.ofile('spanning.alignments.anno', axes), sch.ofile('spanning.alignments.eval', axes))
    sch.transform('eval_split_align', axes, medmem, eval_split_align, None, sch.ifile('simulated.info', axes), sch.ifile('reads1', axes), sch.ifile('split.alignments', axes), sch.ofile('split.alignments.anno', axes), sch.ofile('split.alignments.eval', axes))

    sch.commandline('cluster', axes, himem, cfg.clustermatepairs_tool, '-a', sch.ifile('spanning.alignments', axes), '-m', '1', '-u', sch.iobj('stats', axes).prop('fragment_length_mean'), '-d', sch.iobj('stats', axes).prop('fragment_length_stddev'), '-o', sch.ofile('clusters', axes))
    sch.commandline('breaks', axes, himem, cfg.predictbreaks_tool, '-s', sch.ifile('split.alignments', axes), '-c', sch.ifile('clusters', axes), '-r', cfg.genome_fasta, '-b', sch.ofile('breakpoints', axes))

    sch.transform('results', axes, lowmem, compile_results, None, sch.ifile('simulated.info', axes), sch.ifile('breakpoints', axes), sch.ifile('clusters', axes), sch.ifile('spanning.alignments.eval', axes), sch.ifile('split.alignments.eval', axes), sch.ofile('identified', axes), sch.ofile('classified', axes))

    sch.transform('collate', (), lowmem, collate_results, None, sch.iobj('simulation.params', axes), sch.ifile('identified', axes), sch.ifile('classified', axes), sch.output(args.results))

    pyp.run()

else:


    def create_simulation_params(config_filename):
        config = ConfigParser.ConfigParser()
        config.read(config_filename)
        section = 'main'
        sim_info_base = dict()
        sim_info_base['adjacent_length'] = config.get(section, 'adjacent_length')
        sim_info_base['breakpoints_seed'] = config.get(section, 'breakpoints_seed')
        sim_info_base['dwgsim_seed'] = config.get(section, 'dwgsim_seed')
        sim_info_base['num_breakpoints'] = config.get(section, 'num_breakpoints')
        sim_info_base['num_concordant'] = config.get(section, 'num_concordant')
        read_length_list = config.get(section, 'read_length_list').split()
        fragment_mean_list = config.get(section, 'fragment_mean_list').split()
        fragment_stddev_list = config.get(section, 'fragment_stddev_list').split()
        coverage_list = config.get(section, 'coverage_list').split()
        num_inserted_list = config.get(section, 'num_inserted_list').split()
        homology_list = config.get(section, 'homology_list').split()
        sim_id = 0
        sim_params = dict()
        for read_length, fragment_mean, fragment_stddev, coverage, num_inserted, homology in itertools.product(read_length_list, fragment_mean_list, fragment_stddev_list, coverage_list, num_inserted_list, homology_list):
            sim_params[sim_id] = dict(sim_info_base)
            sim_params[sim_id]['read_length'] = read_length
            sim_params[sim_id]['fragment_mean'] = fragment_mean
            sim_params[sim_id]['fragment_stddev'] = fragment_stddev
            sim_params[sim_id]['coverage'] = coverage
            sim_params[sim_id]['num_inserted'] = num_inserted
            sim_params[sim_id]['homology'] = homology
            sim_id += 1
        return sim_params


    def create_sim_read_table(reads_filename):
        sim_reads = list()
        with open(reads_filename, 'r') as reads_file:
                for name, seq, comment, qual in itertools.izip_longest(*[reads_file]*4):
                        assert name[0] == '@'
                        read_id = int(name[1:].split('/')[0])
                        assert comment[:2] == '+@'
                        sim_id = int(comment[2:].split('_')[0])
                        sim_reads.append((sim_id, read_id))
        sim_reads = pd.DataFrame(sim_reads, columns=['sim_id', 'read_id'])
        return sim_reads


    def read_simulated_table(simulated_filename):
        simulated = pd.read_csv(simulated_filename, sep='\t', header=None,
                                converters={'chromosome_1':str, 'chromosome_2':str},
                                names=['sim_id',
                                       'chromosome_1', 'strand_1', 'position_1',
                                       'chromosome_2', 'strand_2', 'position_2',
                                       'inserted', 'homology'])
        return simulated


    def eval_span_align(simulated_filename, reads_filename, spanning_filename, spanning_annotated_filename, eval_filename):

        max_diff = 1000

        sim_reads = create_sim_read_table(reads_filename)
        simulated = read_simulated_table(simulated_filename)

        spanning = pd.read_csv(spanning_filename, sep='\t', header=None, index_col=False,
                               converters={'chromosome':str},
                               names=['read_id', 'read_end',
                                      'chromosome', 'strand', 'start', 'end',
                                      'read_length', 'matched_length', 'score',
                                      'p_align', 'p_chrimeric', 'p_valid'])

        assert len(spanning.index) > 0

        spanning['approx_break'] = (spanning['strand'] == '+') * spanning['end'] + (spanning['strand'] != '+') * spanning['start']

        spanning.set_index(['read_id', 'read_end'], inplace=True)
        spanning['max_p_align'] = spanning.groupby(level=[0,1])['p_align'].max()
        spanning.reset_index(inplace=True)
        spanning['is_best'] = spanning['max_p_align'] == spanning['p_align']

        spanning = spanning.merge(sim_reads, left_on='read_id', right_on='read_id')
        spanning = spanning.merge(simulated, left_on='sim_id', right_on='sim_id')

        for side in ('1', '2'):
            spanning['is_side_'+side] = (spanning['chromosome'] == spanning['chromosome_'+side]) & \
                                        (spanning['strand'] == spanning['strand_'+side]) & \
                                        (spanning['approx_break'] - max_diff <= spanning['position_'+side]) & \
                                        (spanning['approx_break'] + max_diff >= spanning['position_'+side])

        spanning['is_true'] = spanning['is_side_1'] | spanning['is_side_2']
        spanning['is_true_best'] = spanning['is_true'] & spanning['is_best']

        spanning.to_csv(spanning_annotated_filename, sep='\t', index=False)

        spanning = spanning.set_index(['sim_id', 'read_id', 'read_end'])

        true_alignment = spanning['is_true'].groupby(level=[0,1,2]).apply(any).unstack().apply(any, axis=1)
        true_alignment = true_alignment * 1
        true_alignment = true_alignment.groupby(level=0).mean()

        true_and_best_alignment = spanning['is_true_best'].groupby(level=[0,1,2]).apply(any).unstack().apply(any, axis=1)
        true_and_best_alignment = true_and_best_alignment * 1
        true_and_best_alignment = true_and_best_alignment.groupby(level=0).mean()

        span_eval = pd.concat([true_alignment, true_and_best_alignment], axis=1, keys=['true_align', 'true_best_align'])

        span_eval.to_csv(eval_filename, sep='\t', index=True)


    def eval_split_align(simulated_filename, reads_filename, split_filename, split_annotated_filename, eval_filename):

        sim_reads = create_sim_read_table(reads_filename)
        simulated = read_simulated_table(simulated_filename)

        split = pd.read_csv(split_filename, sep='\t', header=None, index_col=False,
                            converters={'chromosome_1':str, 'chromosome_2':str},
                            names=['read_id', 'read_end',
                                   'chromosome_1', 'strand_1', 'position_1',
                                   'chromosome_2', 'strand_2', 'position_2',
                                   'inserted', 'length_1', 'length_2',
                                   'score_1', 'score_2', 'score', 'p_align'])

        split.set_index(['read_id', 'read_end'], inplace=True)
        split['max_p_align'] = split.groupby(level=[0,1])['p_align'].max()
        split.reset_index(inplace=True)
        split['is_best'] = split['max_p_align'] == split['p_align']

        split = split.merge(sim_reads, left_on='read_id', right_on='read_id')

        merge_columns = ['sim_id']
        for side in ('1', '2'):
            merge_columns.append('chromosome_'+side)
            merge_columns.append('strand_'+side)
            merge_columns.append('position_'+side)

        sim_side_1 = simulated[merge_columns]
        sim_side_1['is_side_1'] = True

        split = split.merge(sim_side_1, left_on=merge_columns, right_on=merge_columns, how='left')

        def flip_side(col):
            if col.endswith('_2'):
                return col[:-2]+'_1'
            elif col.endswith('_1'):
                return col[:-2]+'_2'
            else:
                return col

        sim_side_2 = simulated[merge_columns]
        sim_side_2 = sim_side_2.rename(columns=flip_side)
        sim_side_2['is_side_2'] = True

        split = split.merge(sim_side_2, left_on=merge_columns, right_on=merge_columns, how='left')

        split['is_side_1'] = split['is_side_1'].fillna(False)
        split['is_side_2'] = split['is_side_2'].fillna(False)

        split['is_true'] = split['is_side_1'] | split['is_side_2']
        split['is_true_best'] = split['is_true'] & split['is_best']

        split.to_csv(split_annotated_filename, sep='\t', index=False)

        split = split.set_index(['sim_id', 'read_id', 'read_end'])

        true_alignment = split['is_true'].groupby(level=[0,1,2]).apply(any)
        true_alignment = true_alignment * 1
        true_alignment = true_alignment.groupby(level=0).mean()

        true_and_best_alignment = split['is_true_best'].groupby(level=[0,1,2]).apply(any)
        true_and_best_alignment = true_and_best_alignment * 1
        true_and_best_alignment = true_and_best_alignment.groupby(level=0).mean()

        split_eval = pd.concat([true_alignment, true_and_best_alignment], axis=1, keys=['true_align', 'true_best_align'])

        split_eval.to_csv(eval_filename, sep='\t', index=True)


    def collate_results(sim_params, identified_filenames, classified_filenames, results_filename):

        with open(results_filename, 'w') as results_file:

            fields = list(set([field for sim in sim_params.values() for field in sim.keys()]))
            
            results_file.write('\t'.join(['sim_id'] + fields + ['sensitivity_exact', 'sensitivity_approx', 'homology_sensitivity', 'specificity', 'read_count', 'span_true_align', 'span_true_best_align', 'split_true_align', 'split_true_best_align']) + '\n')
            
            for sim_id in sim_params.keys():

                results_file.write(str(sim_id) + '\t')
                results_file.write('\t'.join((sim_params[sim_id][field] for field in fields)) + '\t')

                identified = pd.read_csv(identified_filenames[sim_id], sep='\t')
                sensitivity_exact = identified['exact_cluster_id'].notnull().mean()
                sensitivity_approx = identified['approx_cluster_id'].notnull().mean()
                homology_sensitivity = identified['homology_correct'].mean()
                read_count = identified['read_count'].mean()
                span_true_align = identified['span_true_align'].mean()
                span_true_best_align = identified['span_true_best_align'].mean()
                split_true_align = identified['split_true_align'].mean()
                split_true_best_align = identified['split_true_best_align'].mean()

                classified = pd.read_csv(classified_filenames[sim_id], sep='\t')
                specificity = classified['is_true'].mean()

                results_file.write(str(sensitivity_exact) + '\t')
                results_file.write(str(sensitivity_approx) + '\t')
                results_file.write(str(homology_sensitivity) + '\t')
                results_file.write(str(specificity) + '\t')
                results_file.write(str(read_count) + '\t')
                results_file.write(str(span_true_align) + '\t')
                results_file.write(str(span_true_best_align) + '\t')
                results_file.write(str(split_true_align) + '\t')
                results_file.write(str(split_true_best_align) + '\n')


    def read_sim(input_file):
        for row in csv.reader(input_file, delimiter='\t'):
            sim_id = int(row[0])
            breakends = frozenset([(row[1],row[2],int(row[3])),(row[4],row[5],int(row[6]))])
            homology = int(row[8])
            yield sim_id, breakends, homology


    def read_breakends(input_file):
        prediction_id = 0
        for row in csv.reader(input_file, delimiter='\t'):
            cluster_id = int(row[0])
            breakends = frozenset([(row[1],row[2],int(row[3])),(row[4],row[5],int(row[6]))])
            yield prediction_id, cluster_id, breakends
            prediction_id += 1


    def read_homology(input_file):
        for cluster_id, rows in itertools.groupby(csv.reader(input_file, delimiter='\t'), lambda row: int(row[0])):
            cluster_id = int(cluster_id)
            rows = list(rows)
            homology = 0
            prev_pos = None
            prev_dist = None
            for row in sorted(rows, key=lambda row: int(row[3])):
                current_pos = int(row[3])
                if row[2] == row[5]:
                    current_dist = int(row[3]) + int(row[6])
                else:
                    current_dist = int(row[3]) - int(row[6])
                if prev_pos is not None:
                    if current_pos == prev_pos + 1 and current_dist == prev_dist:
                        homology += 1
                    else:
                        homology = 0
                prev_pos = current_pos
                prev_dist = current_dist
            yield cluster_id, homology


    def read_cluster_regions(input_file):
        for cluster_id, rows in itertools.groupby(csv.reader(input_file, delimiter='\t'), lambda row: row[0]):
            cluster_id = int(cluster_id)
            region = [[None, None, [], []], [None, None, [], []]]
            read_ids = set()
            for row in rows:
                cluster_end = int(row[1])
                region[cluster_end][0] = row[4]
                region[cluster_end][1] = row[5]
                region[cluster_end][2].append(int(row[6]))
                region[cluster_end][3].append(int(row[7]))
                read_ids.add(row[2])
            for cluster_end in (0, 1):
                region[cluster_end][2] = min(region[cluster_end][2])
                region[cluster_end][3] = max(region[cluster_end][3])
            yield cluster_id, region, len(read_ids)


    def match(region, breakpoint):
        min_dist = 200
        for side in (0, 1):
            if region[0][0] != breakpoint[side][0] or region[0][1] != breakpoint[side][1] or region[1][0] != breakpoint[1-side][0] or region[1][1] != breakpoint[1-side][1]:
                continue
            if breakpoint[side][2] > region[0][3] + min_dist or breakpoint[side][2] < region[0][2] - min_dist or breakpoint[1-side][2] > region[1][3] + min_dist or breakpoint[1-side][2] < region[1][2] - min_dist:
                continue
            return True
        return False


    def compile_results(sim_filename, breakpoints_filename, clusters_filename, span_eval_filename, split_eval_filename, identified_filename, classified_filename):

        predictions = collections.defaultdict(list)
        with open(breakpoints_filename, 'r') as breakpoints_file:
            for prediction_id, cluster_id, breakends in read_breakends(breakpoints_file):
                predictions[breakends].append(cluster_id)
        
        predicted_homology = dict()
        with open(breakpoints_filename, 'r') as breakpoints_file:
            for cluster_id, homology in read_homology(breakpoints_file):
                predicted_homology[cluster_id] = homology
        
        cluster_regions = dict()
        cluster_read_counts = dict()
        with open(clusters_filename, 'r') as clusters_file:
            for cluster_id, region, read_count in read_cluster_regions(clusters_file):
                cluster_regions[cluster_id] = region
                cluster_read_counts[cluster_id] = read_count
        
        span_eval = pd.read_csv(span_eval_filename, sep='\t', index_col='sim_id')
        
        split_eval = pd.read_csv(split_eval_filename, sep='\t', index_col='sim_id')
        
        with open(sim_filename, 'r') as sim_file, open(identified_filename, 'w') as identified_file, open(classified_filename, 'w') as classified_file:

            true_cluster_ids = set()
            identified_file.write('\t'.join(['sim_id', 'exact_cluster_id', 'approx_cluster_id', 'homology_correct', 'read_count', 'span_true_align', 'span_true_best_align', 'split_true_align', 'split_true_best_align']) + '\n')

            for sim_id, breakends, homology in read_sim(sim_file):

                exact_cluster_id = 'NA'
                approx_cluster_id = 'NA'
                homology_correct = '0'
                for cluster_id in predictions[breakends]:
                    exact_cluster_id = cluster_id
                    true_cluster_ids.add(cluster_id)
                    if predicted_homology[cluster_id] == homology:
                        homology_correct = '1'
                for cluster_id, region in cluster_regions.items():
                    if match(region, list(breakends)):
                        approx_cluster_id = cluster_id
                        true_cluster_ids.add(approx_cluster_id)

                read_count = 'NA'
                if exact_cluster_id != 'NA':
                    read_count = cluster_read_counts[exact_cluster_id]
                elif approx_cluster_id != 'NA':
                    read_count = cluster_read_counts[approx_cluster_id]

                identified_file.write(str(sim_id) + '\t')
                identified_file.write(str(exact_cluster_id) + '\t')
                identified_file.write(str(approx_cluster_id) + '\t')
                identified_file.write(str(homology_correct) + '\t')
                identified_file.write(str(read_count) + '\t')
                identified_file.write(str(span_eval['true_align'].get(sim_id, default=0.0)) + '\t')
                identified_file.write(str(span_eval['true_best_align'].get(sim_id, default=0.0)) + '\t')
                identified_file.write(str(split_eval['true_align'].get(sim_id, default=0.0)) + '\t')
                identified_file.write(str(split_eval['true_best_align'].get(sim_id, default=0.0)) + '\n')
            
            classified_file.write('\t'.join(['cluster_id', 'is_true']) + '\n')
            for cluster_id, region in cluster_regions.items():
                classified_file.write('\t'.join([str(cluster_id), str(int(cluster_id in true_cluster_ids))]) + '\n')


    class ConcordantReadStatsAccumulator(object):
        def __init__(self, fragment_length_num_stddevs):
            self.fragment_length_num_stddevs = fragment_length_num_stddevs
            self.read_lengths = set()
            self.fragment_length_sum = 0
            self.fragment_length_sum_sq = 0
            self.fragment_count = 0
        def add(self, read_length_1, read_length_2, fragment_length):
            self.read_lengths.add(read_length_1)
            self.read_lengths.add(read_length_2)
            self.fragment_length_sum += float(fragment_length)
            self.fragment_length_sum_sq += float(fragment_length) ** 2
            self.fragment_count += 1
        def update(self, other):
            self.read_lengths.update(other.read_lengths)
            self.fragment_length_sum += other.fragment_length_sum
            self.fragment_length_sum_sq += other.fragment_length_sum_sq
            self.fragment_count += other.fragment_count
        @property
        def fragment_length_mean(self):
            return self.fragment_length_sum / self.fragment_count
        @property
        def fragment_length_stddev(self):
            return self.fragment_length_variance ** 0.5
        @property
        def fragment_length_variance(self):
            return self.fragment_length_sum_sq / self.fragment_count - self.fragment_length_mean ** 2
        @property
        def fragment_length_min(self):
            return int(max(0, self.fragment_length_mean - self.fragment_length_num_stddevs * self.fragment_length_stddev))
        @property
        def fragment_length_max(self):
            return int(self.fragment_length_mean + self.fragment_length_num_stddevs * self.fragment_length_stddev)


    class SamRecord(object):
        def __init__(self, line):
            self.line = line
            self.fields = line.rstrip().split('\t')
            self.fields[1] = int(self.fields[1])
        @property
        def is_read_mapped(self):
            return self.fields[1] & 0x04 == 0
        @property
        def is_pair_mapped(self):
            return self.fields[1] & 0x02 != 0
        @property
        def read_id(self):
            return self.fields[0]
        @property
        def fragment_id(self):
            return self.fields[0]
        @property
        def read_end(self):
            if self.fields[1] & 0x40:
                return 0
            elif self.fields[1] & 0x80:
                return 1
            else:
                return None
        @property
        def read_length(self):
            return len(self.fields[9])
        @property
        def chromosome(self):
            return self.fields[2]
        @property
        def position(self):
            return int(self.fields[3])
        @property
        def strand(self):
            if self.fields[1] & 0x10:
                return '-'
            else:
                return '+'
        @property
        def isize(self):
            return abs(int(self.fields[8]))
        @property
        def alignment_score(self):
            for opt in self.fields[11:]:
                if opt.startswith('AS:i:'):
                    return int(opt[5:])
        @property
        def num_matches(self):
            return self.matched - self.num_mismatches
        @property
        def num_mismatches(self):
            for opt in self.fields[11:]:
                if opt.startswith('XM:i:'):
                    return int(opt[5:])
        @property
        def start_gap(self):
            start_gap_match = self.start_gap_re.search(self.fields[5])
            if start_gap_match is not None:
                return int(start_gap_match.group(1))
            else:
                return 0
        start_gap_re = re.compile('^(\d+)S')
        @property
        def end_gap(self):
            end_gap_match = self.end_gap_re.search(self.fields[5])
            if end_gap_match is not None:
                return int(end_gap_match.group(1))
            else:
                return 0
        end_gap_re = re.compile('(\d+)S$')
        @property
        def matched(self):
            matched = 0
            for count in self.matched_re.findall(self.fields[5]):
                matched += int(count)
            return matched
        matched_re = re.compile('(\d+)M')
        @property
        def deleted(self):
            deleted = 0
            for count in self.deleted_re.findall(self.fields[5]):
                deleted += int(count)
            return deleted
        deleted_re = re.compile('(\d+)D')


    def calculate_concordant_stats(concordant_sam, fragment_length_num_stddevs):
        stats = ConcordantReadStatsAccumulator(fragment_length_num_stddevs)
        try:
            with open(concordant_sam, 'r') as concordant:
                for line_1 in concordant:
                    if line_1[0] == '@':
                        continue
                    line_2 = concordant.next()
                    record_1 = SamRecord(line_1)
                    record_2 = SamRecord(line_2)
                    if (record_1.fragment_id != record_2.fragment_id):
                        logging.getLogger('pipeline').error('Mismatched read names %s and %s' % (record_1.read_id, record_2.read_id))
                    if record_1.isize != record_2.isize:
                        logging.getLogger('pipeline').error('Mismatched isizes %d and %d for reads %s and %s' % (record_1.isize, record_2.isize, record_1.read_id, record_2.read_id))
                    if record_1.chromosome != record_2.chromosome:
                        logging.getLogger('pipeline').error('Mismatched references for concordant reads %s and %s' % (record_1.read_id, record_2.read_id))
                    if not record_1.is_pair_mapped or not record_2.is_pair_mapped:
                        continue
                    stats.add(record_1.read_length, record_2.read_length, record_1.isize)
        except StopIteration:
            pass
        return stats


    def index_fastq(in_filename, out_filename):
        with open(in_filename, 'r') as in_file, open(out_filename, 'w') as out_file:
            read_index = 0
            for name, seq, comment, qual in itertools.izip_longest(*[in_file]*4):
                out_file.write('@' + str(read_index) + name.rstrip()[-2:] + '\n')
                out_file.write(seq)
                out_file.write('+' + name)
                out_file.write(qual)
                read_index += 1



import string
import numpy as np


def reverse_complement(sequence):
    return sequence[::-1].translate(str.maketrans('ACTGactg','TGACtgac'))

def column_flip(df, cond, col_1, col_2):
    df.loc[cond, col_1], df.loc[cond, col_2] = df.loc[cond, col_2], df.loc[cond, col_1]


def calculate_offset(strand, offset):
    if (strand == "+"):
        direction = 1
    else:
        direction = -1
    return offset * direction


def calculate_forward_homology(chromosome, strand, position, genome, maxOffset, flip=False):
    if flip:
        idx1 = 1
    else:
        idx1 = 0
    idx2 = 1 - idx1

    chr1 = genome[chromosome[idx1]]
    chr2 = genome[chromosome[idx2]]

    homology = 0
    for offset in range(1, maxOffset + 1):
        nt1 = chr1[position[idx1] + calculate_offset(strand[idx1], offset) - 1]
        nt2 = chr2[position[idx2] + calculate_offset(strand[idx2], 1 - offset) - 1]

        if (strand[idx1] != '+'):
            nt1 = reverse_complement(nt1)
        
        if (strand[idx2] != '-'):
            nt2 = reverse_complement(nt2)

        if (nt1 != nt2):
            break
        
        homology = offset
    
    return homology


def homology_consistent_breakpoint(chromosome, strand, position, genome, maxOffset):
    maxOffsetA = calculate_forward_homology(chromosome, strand, position, genome, maxOffset, False)
    maxOffsetB = calculate_forward_homology(chromosome, strand, position, genome, maxOffset, True)

    # Ensure that the same breakpoint is selected among the multiple
    # breakpoints possible when there is breakpoint homology.  Always
    # select the breakpoint for which the minimum of the two breakend
    # positions is minimal

    positionA1 = position[0] + calculate_offset(strand[0], maxOffsetA)
    positionA2 = position[1] + calculate_offset(strand[1], -maxOffsetA)

    positionB1 = position[0] + calculate_offset(strand[0], -maxOffsetB)
    positionB2 = position[1] + calculate_offset(strand[1], maxOffsetB)

    if (min(positionA1, positionA2) < min(positionB1, positionB2)):
        position[0] = positionA1
        position[1] = positionA2
    else:
        position[0] = positionB1
        position[1] = positionB2

    homology = maxOffsetA + maxOffsetB

    return homology


def normalize_breakpoint(chr1, str1, pos1, chr2, str2, pos2, genome, max_offset=100):
    chromosome = [chr1, chr2]
    strand = [str1, str2]
    position = [pos1, pos2]

    homology = homology_consistent_breakpoint(
        chromosome, strand, position, genome, max_offset)

    return position[0], position[1], homology




import argparse

import pypeliner

import destruct
import destruct.workflow


def run(**args):
    if len(args['bam_files']) != len(args['lib_ids']):
        raise Exception('--lib_ids must correspond one to one with --bam_files')

    bam_filenames = dict(zip(args['lib_ids'], args['bam_files']))

    config = {}
    if args['config'] is not None:
        exec(open(args['config']).read(), {}, config)

    pyp = pypeliner.app.Pypeline(modules=[destruct], config=args)

    workflow = destruct.workflow.create_destruct_workflow(
        bam_filenames,
        args['breakpoint_table'],
        args['breakpoint_library_table'],
        args['breakpoint_read_table'],
        config,
        args['ref_data_dir'],
        args['raw_data_dir'],
    )

    pyp.run(workflow)


def add_arguments(argparser):
    pypeliner.app.add_arguments(argparser)

    argparser.add_argument('ref_data_dir',
                           help='Reference dataset directory')

    argparser.add_argument('breakpoint_table',
                           help='Output table of breakpoint information in TSV format')

    argparser.add_argument('breakpoint_library_table',
                           help='Output table of library specific breakpoint information in TSV format')

    argparser.add_argument('breakpoint_read_table',
                           help='Output table of breakpoint read information in TSV format')

    argparser.add_argument('--bam_files', nargs='+',
                           help='Input bam filenames')

    argparser.add_argument('--lib_ids', nargs='+',
                           help='Input ids for respective bam filenames')

    argparser.add_argument('--config', required=False,
                           help='Configuration filename')

    argparser.add_argument('--raw_data_dir', required=False,
                           help='Raw data directory')

    argparser.set_defaults(func=run)


if __name__ == '__main__':
    argparser = argparse.ArgumentParser()

    add_arguments(argparser)

    args = vars(argparser.parse_args())
    func = args.pop('func')
    func(**args)
